import copy
from typing import List, Dict
import pronto
from .. import utils
from ..io import iobase
from ..orm import general, cv


class OntologyClient(iobase.IOClient):

    def __init__(self, uri: str, verbose=False):
        """Constructor"""

        # Connect to database
        super().__init__(uri)

        # Set up printer
        self.printer = utils.VerbosePrinter(verbose)
        
        # Initiate counters
        self._db_inserts = 0
        self._cv_inserts = 0
        self._dbxref_inserts = 0
        self._dbxref_updates = 0
        self._cvterm_inserts = 0
        self._cvterm_updates = 0
        self._cvterm_deletes = 0
        self._comment_inserts = 0
        self._comment_updates = 0
        self._comment_deletes = 0
        self._synonym_inserts = 0
        self._synonym_updates = 0
        self._synonym_deletes = 0
        self._crossref_inserts = 0
        self._crossref_updates = 0
        self._crossref_deletes = 0
        self._relationship_inserts = 0
        self._relationship_updates = 0
        self._relationship_deletes = 0

        # Load essentials
        self._comment_term = self._load_comment_term()
        self._relationship_terms = self._load_relationship_terms()
        self._synonym_type_terms = self._load_synonym_type_terms()

    def load(self, filename: str, file_format: str, db_authority: str):
        """Loads CV terms from a file into a database"""

        # Parse the file
        self.printer.print("Parsing ontology file ...")
        ontology = parse_ontology(filename, file_format)                            # type: pronto.Ontology

        # Filter content: Only retain the terms stemming from the database authority of interest
        default_namespace = get_default_namespace(ontology)
        ontology_terms = filter_ontology_by_db(ontology, db_authority)              # type: Dict[str, pronto.Term]
        self.printer.print("Retrieved " + str(len(ontology_terms)) + " terms for database authority " + db_authority)

        # Find/create parent vocabulary and db of ontology terms in file; load dependencies
        default_db_entry = self._handle_db(db_authority)
        default_cv_entry = self._handle_cv(pronto.Term(""), default_namespace)

        # Create global containers to reduce the number of database requests
        all_cvterm_entries = {}                                                     # type: Dict[str, cv.CvTerm]

        # Handle typedefs
        for typedef in ontology.typedefs:                                           # type: pronto.Relationship
            self._handle_typedef(typedef, default_db_entry, default_cv_entry)

        # First loop over all terms retrieved from the file: handle vocabulary, dbxref, CV terms, synonyms, comments
        for term in ontology_terms.values():                                        # type: pronto.Term

            # Insert, update and/or delete entries in various tables
            cv_entry = self._handle_cv(term, default_namespace)
            dbxref_entry = self._handle_dbxref(term, default_db_entry)
            cvterm_entry = self._handle_cvterms(ontology_terms, default_db_entry, dbxref_entry, cv_entry.cv_id)
            self._handle_comments(term, cvterm_entry)
            self._handle_synonyms(term, cvterm_entry)

            # Save CV term in global container
            all_cvterm_entries[term.id] = cvterm_entry

        # Second loop over all terms retrieved from file: cross references and relationships
        for term in ontology_terms.values():                                        # type: pronto.Term

            # Insert, update and/or delete entries in various tables
            self._handle_cross_references(term, all_cvterm_entries[term.id])
            self._handle_relationships(term, all_cvterm_entries[term.id], all_cvterm_entries)

        # Mark obsolete CV terms
        self._mark_obsolete_terms(ontology_terms, default_db_entry)

        # Commit changes
        self.session.commit()
        self._print_statistics()

    def _load_synonym_type_terms(self) -> Dict[str, cv.CvTerm]:
        """Loads CV terms describing tyoes of synonyms"""
        synonym_type_cv = self.query_table(cv.Cv, name="synonym_type").first()
        if not synonym_type_cv:
            raise iobase.DatabaseError("CV 'synonym_type' not present in database")
        synonym_type_cvterms = self.query_table(cv.CvTerm, cv_id=synonym_type_cv.cv_id).all()
        synonym_type_cvterms_dict = utils.list_to_dict(synonym_type_cvterms, "name")
        required_terms = ["exact", "narrow", "broad", "related"]
        for term in required_terms:
            if term not in synonym_type_cvterms_dict:
                raise iobase.DatabaseError("CV term for synonym type '" + term + "' not present in database")
        return synonym_type_cvterms_dict

    def _load_comment_term(self) -> cv.CvTerm:
        """Loads the CV term describing a comment"""
        comment_cvterm = self.query_table(cv.CvTerm, name="comment").first()
        if not comment_cvterm:
            raise iobase.DatabaseError("CV term 'comment' not present in database")
        return comment_cvterm

    def _load_relationship_terms(self) -> Dict[str, cv.CvTerm]:
        """Loads CV terms describing relationships between CV terms"""
        relationship_cv = self.query_table(cv.Cv, name="relationship").first()
        if not relationship_cv:
            raise iobase.DatabaseError("CV 'relationship' not present in database")
        relationship_cvterms = self.query_table(cv.CvTerm, is_relationshiptype=1).all()
        relationship_cvterms_dict = utils.list_to_dict(relationship_cvterms, "name")
        required_terms = ["is_a", "part_of"]
        for term in required_terms:
            if term not in relationship_cvterms_dict:
                raise iobase.DatabaseError("CV term for relationship '" + term + "' not present in database")
        return relationship_cvterms_dict

    def _handle_db(self, db_authority: str) -> general.Db:
        """Returns an entry from the db table"""
        db_entry = self.query_table(general.Db, name=db_authority).first()         # type: general.Db
        if not db_entry:
            db_entry = general.Db(name=db_authority)
            self.add_and_flush(db_entry)
            self.printer.print("Inserted DB '" + db_authority + "'")
            self._db_inserts += 1
        return db_entry

    def _handle_cv(self, term: pronto.Term, default_namespace: str) -> cv.Cv:
        """Inserts a controlled vocabulary in the database, and returns the entry"""

        # Get the namespace of the term from the input file
        namespace = default_namespace
        if "namespace" in term.other:
            namespace = term.other["namespace"][0]
        if not namespace:
            raise iobase.InputFileError("Namespace missing in input file")

        # Get the corresponding CV in the database - create it, if not yet available
        cv_entry = self.query_table(cv.Cv, name=namespace).first()                 # type: cv.Cv
        if not cv_entry:
            cv_entry = cv.Cv(name=namespace)
            self.add_and_flush(cv_entry)
            self.printer.print("Inserted CV '" + namespace + "'")
            self._cv_inserts += 1
        return cv_entry

    def _handle_dbxref(self, term: pronto.Term, default_db: general.Db) -> general.DbxRef:
        """Inserts or updates a database cross reference in the dbxref table, and returns the entry"""

        # Split database cross reference (dbxref) into db, accession, version
        (db_authority, accession, version) = split_dbxref(term.id)

        # Check if the dbxref is already present in the database
        existing_dbxref_entries = self.query_table(general.DbxRef, db_id=default_db.db_id,
                                                   accession=accession).all()          # type: List[general.DbxRef]
        if existing_dbxref_entries:

            # Check if the entries in database and file are identical
            for dbxref_entry in existing_dbxref_entries:                                # type: general.DbxRef
                if dbxref_entry.version == version:
                    break
            else:

                # Update dbxref (version)
                dbxref_entry = existing_dbxref_entries[0]                               # type: general.DbxRef
                dbxref_entry.version = version
                self.printer.print("Updated dbxref '" + term.id + "'")
                self._dbxref_updates += 1
        else:

            # Insert dbxref
            dbxref_entry = general.DbxRef(db_id=default_db.db_id, accession=accession, version=version)
            self.add_and_flush(dbxref_entry)
            self.printer.print("Inserted dbxref '" + term.id + "'")
            self._dbxref_inserts += 1

        # Return the entry
        return dbxref_entry

    def _handle_cvterms(self, ontology_terms: Dict[str, pronto.Term], db_entry: general.Db,
                        dbxref_entry: general.DbxRef, cv_id: int) -> cv.CvTerm:
        """Inserts or updates a CV term in the cvterm table, and makes sure not to violate any UNIQUE constraints"""

        # Check if there is an entry in the input ontology
        dbxref = create_dbxref(db_entry.name, dbxref_entry.accession, dbxref_entry.version)
        if dbxref in ontology_terms:

            # Create a CV term and check if it is unique (if inserted into the database)
            term = ontology_terms[dbxref]                                                   # type: pronto.Term
            cvterm_entry = create_cvterm_entry(term, cv_id, dbxref_entry.dbxref_id)
            if cvterm_entry.is_obsolete:
                self._mark_cvterm_as_obsolete(cvterm_entry)
            (is_unique, existing_cvterm_entry, existing_dbxref_entry) \
                = self._is_cvterm_unique(cvterm_entry)                   # type: bool, cv.CvTerm, general.DbxRef
            if not is_unique:

                # Recursive call to update the existing CV term first
                self._handle_cvterms(ontology_terms, db_entry, existing_dbxref_entry, existing_cvterm_entry.cv_id)

            if is_unique and existing_cvterm_entry:

                # There is a CV term in the database that coincides with the input in terms of cv, name,
                # is_obsolete and dbxref_id. Update if necessary.
                if update_cvterm_properties(existing_cvterm_entry, cvterm_entry):
                    self.printer.print("Updated CV term '" + existing_cvterm_entry.name + "' for dbxref " + term.id)
                    self._cvterm_updates += 1
                return existing_cvterm_entry
            else:

                # There is no CV term in the database that coincides with the input in terms of cv,
                # name and is_obsolete. Check if there is one with the same dbxref.
                existing_cvterm_entry = self.query_table(cv.CvTerm, dbxref_id=cvterm_entry.dbxref_id).first()
                if existing_cvterm_entry:

                    # Update parameters, if necessary
                    if update_cvterm_properties(existing_cvterm_entry, cvterm_entry):
                        self.printer.print("Updated CV term '" + existing_cvterm_entry.name + "' for dbxref " + term.id)
                        self._cvterm_updates += 1
                    return existing_cvterm_entry
                else:

                    # Insert a new CV term
                    self.add_and_flush(cvterm_entry)
                    self.printer.print("Inserted CV term '" + cvterm_entry.name + "' for dbxref " + term.id)
                    self._cvterm_inserts += 1
                    return cvterm_entry
        else:

            # Mark the CV term as obsolete (or "more obsolete")
            existing_cvterm_entry = self.query_table(cv.CvTerm, dbxref_id=dbxref_entry.dbxref_id).first()
            if self._mark_cvterm_as_obsolete(existing_cvterm_entry):
                self.printer.print("Marked CV term '" + existing_cvterm_entry.name + "' as obsolete")
                self._cvterm_deletes += 1
            return existing_cvterm_entry

    def _is_cvterm_unique(self, cvterm_entry: cv.CvTerm) -> (bool, cv.CvTerm, general.DbxRef):
        """Checks if a CV term would fulfill all UNIQUE constraints if inserted into the database"""

        # Check if a CV term with the same properties exists in the database
        confounding_cvterm_entry = self.query_table(cv.CvTerm, cv_id=cvterm_entry.cv_id, name=cvterm_entry.name,
                                                    is_obsolete=cvterm_entry.is_obsolete).first()   # type: cv.CvTerm

        if confounding_cvterm_entry and confounding_cvterm_entry.dbxref_id != cvterm_entry.dbxref_id:

            # There is a CV term with these properties in the database, and its ID differs from the input term
            # -> non-unique
            corresponding_debxref_entry = self.query_table(
                general.DbxRef, dbxref_id=confounding_cvterm_entry.dbxref_id).first()   # type: general.DbxRef
            return False, confounding_cvterm_entry, corresponding_debxref_entry

        elif confounding_cvterm_entry:

            # There is a CV term with these properties in the database, but its ID aligns with the input term -> unique
            return True, confounding_cvterm_entry, None

        else:

            # There is no CV term with these properties in the database -> unique
            return True, None, None

    def _mark_cvterm_as_obsolete(self, cvterm_entry: cv.CvTerm) -> bool:
        """Marks a CV term in the database as obsolete, and makes sure this doesn't violate any UNIQUE constraints"""

        # Create a copy of the CV term entry and mark that as obsolete
        marked = False
        test_cvterm_entry = copy.deepcopy(cvterm_entry)
        if "obsolete" not in test_cvterm_entry.name.lower():
            test_cvterm_entry.name = "obsolete " + test_cvterm_entry.name
        test_cvterm_entry.is_obsolete = 1

        # Check if the changed CV term still fulfills all UNIQUE constraints. If not, increase the is_obsolete parameter
        while not self._is_cvterm_unique(test_cvterm_entry)[0]:
            test_cvterm_entry.is_obsolete += 1

        # Transfer the changed properties to the original CV term
        if test_cvterm_entry != cvterm_entry:
            marked = True
            cvterm_entry.name = test_cvterm_entry.name
            cvterm_entry.is_obsolete = test_cvterm_entry.is_obsolete
        return marked

    def _handle_typedef(self, typedef: pronto.Relationship, default_db: general.Db, default_cv: cv.Cv):
        """Inserts CV terms for relationship ontology terms (so-called "typedefs")"""

        # Check if a relationship CV term with this name already exists
        cvterm_entry = self.query_table(cv.CvTerm, name=typedef.obo_name, is_relationshiptype=1).first()
        if not cvterm_entry:

            # Create dbxref
            dbxref = create_dbxref(default_db.name, typedef.obo_name)
            term = pronto.Term(dbxref, typedef.obo_name)
            dbxref_entry = self._handle_dbxref(term, default_db)

            # Create CV term
            cvterm_entry = cv.CvTerm(cv_id=default_cv.cv_id, dbxref_id=dbxref_entry.dbxref_id,
                                     name=typedef.obo_name, is_relationshiptype=1)
            self.add_and_flush(cvterm_entry)
            self._cvterm_inserts += 1
            self.printer.print("Inserted CV term '" + cvterm_entry.name + "' for dbxref " + dbxref)
        return cvterm_entry

    def _handle_comments(self, term: pronto.Term, cvterm_entry: cv.CvTerm) -> cv.CvTermProp:
        """Inserts, updates or deletes comments to a CV term in the cvtermprop table"""

        # Get existing and new comments, i.e. comments present in database and in input file
        existing_comment_entries = self.query_table(cv.CvTermProp, cvterm_id=cvterm_entry.cvterm_id,
                                                    type_id=self._comment_term.cvterm_id).all()
        comment = extract_comment(term)
        edited_entry = None

        # Update/insert comment, if available
        if comment:

            # Check if there is already a comment to this CV term in the database
            if existing_comment_entries:

                # Check if the comments in database and file are identical
                for cvtermprop_entry in existing_comment_entries:                   # type: cv.CvTermProp
                    if cvtermprop_entry.value == comment:
                        break
                else:

                    # Update comment
                    cvtermprop_entry = existing_comment_entries[0]                  # type: cv.CvTermProp
                    cvtermprop_entry.value = comment
                    self.printer.print("Updated comment '" + cvtermprop_entry.value + "' for CV term '"
                                       + cvterm_entry.name + "'")
                    self._comment_updates += 1
            else:

                # Insert comment
                cvtermprop_entry = cv.CvTermProp(cvterm_id=cvterm_entry.cvterm_id, type_id=self._comment_term.cvterm_id,
                                                 value=comment)
                self.add_and_flush(cvtermprop_entry)
                self.printer.print("Inserted comment '" + cvtermprop_entry.value + "' for CV term '"
                                   + cvterm_entry.name + "'")
                self._comment_inserts += 1
            edited_entry = cvtermprop_entry

        # Delete comments not present in file
        for comment_entry in existing_comment_entries:                              # type: cv.CvTermProp
            if not comment_entry.value or comment_entry.value != comment:

                # Delete comment
                self.session.delete(comment_entry)
                self.printer.print("Deleted comment '" + comment_entry.value + "' for CV term '"
                                   + cvterm_entry.name + "'")
                self._comment_deletes += 1

        return edited_entry

    def _handle_synonyms(self, term: pronto.Term, cvterm_entry: cv.CvTerm) -> List[cv.CvTermSynonym]:
        """Inserts, updates or deletes synonyms to a CV term in the cvtermsynonym table"""

        # Get existing and new synonyms, i.e. synonyms present in database and in input file
        existing_synonyms = self.query_table(cv.CvTermSynonym, cvterm_id=cvterm_entry.cvterm_id).all()
        new_synonyms = extract_synonyms(term)
        edited_entries = []

        # Loop over all synonyms in input file
        for synonym in term.synonyms:                                               # type: pronto.Synonym

            # Extract synonym type
            if synonym.scope.lower() not in self._synonym_type_terms:
                synonym_type_term = ""
                self.printer.print("WARNING: synonym type '" + synonym.scope.lower() + "' not present in database!")
            else:
                synonym_type_term = self._synonym_type_terms[synonym.scope.lower()]

            # Check if the synonym is already present in the database
            matching_synonyms = utils.filter_objects(existing_synonyms,
                                                     synonym=synonym.desc)          # type: List[cv.CvTermSynonym]
            if matching_synonyms:

                # Check if the synonyms in database and file have identical properties (len(matching_synonyms) = 1)
                cvtermsynonym_entry = matching_synonyms[0]                          # type: cv.CvTermSynonym
                if cvtermsynonym_entry.type_id != synonym_type_term.cvterm_id:

                    # Update synonym (type)
                    cvtermsynonym_entry.type_id = synonym_type_term.cvterm_id
                    self.printer.print("Updated synonym '" + cvtermsynonym_entry.synonym + "' for CV term '"
                                       + cvterm_entry.name + "'")
                    self._synonym_updates += 1
            else:

                # Insert synonym
                cvtermsynonym_entry = cv.CvTermSynonym(cvterm_id=cvterm_entry.cvterm_id, synonym=synonym.desc,
                                                       type_id=synonym_type_term.cvterm_id)
                self.add_and_flush(cvtermsynonym_entry)
                self.printer.print("Inserted synonym '" + cvtermsynonym_entry.synonym + "' for CV term '"
                                   + cvterm_entry.name + "'")
                self._synonym_inserts += 1

            edited_entries.append(cvtermsynonym_entry)

        # Delete synonyms not present in file
        for cvtermsynonym_entry in existing_synonyms:                               # type: cv.CvTermSynonym
            if not cvtermsynonym_entry.synonym or cvtermsynonym_entry.synonym not in new_synonyms:

                # Delete synonym
                self.session.delete(cvtermsynonym_entry)
                self.printer.print("Deleted synonym '" + cvtermsynonym_entry.synonym + "' for CV term '"
                                   + cvterm_entry.name + "'")
                self._synonym_deletes += 1

        return edited_entries

    def _handle_cross_references(self, term: pronto.Term, cvterm_entry: cv.CvTerm) -> List[cv.CvTermDbxRef]:
        """Inserts, updates or deletes database cross references to a CV term in the cvterm_dbxref table"""

        # Get existing cross references, i.e. cross references present in database
        existing_crossrefs = self.query_table(cv.CvTermDbxRef, cvterm_id=cvterm_entry.cvterm_id).all()
        edited_entries = []

        # Get cross references present in the file and loop over them
        new_crossrefs = extract_cross_references(term)
        for crossref in new_crossrefs:

            # Obtain the corresponding entry from the dbxref table, insert if not existing
            db_authority = split_dbxref(crossref)[0]
            corresponding_db_entry = self._handle_db(db_authority)
            corresponding_dbxref_entry = self._handle_dbxref(pronto.Term(crossref), corresponding_db_entry)

            # Check if the cross reference is already present in the database
            matching_crossrefs = utils.filter_objects(
                existing_crossrefs, dbxref_id=corresponding_dbxref_entry.dbxref_id)  # type: List[cv.CvTermDbxRef]
            if matching_crossrefs:

                # Save the existing cross reference
                edited_entries.append(matching_crossrefs[0])
            else:

                # Insert cross reference
                cvterm_dbxref_entry = cv.CvTermDbxRef(cvterm_id=cvterm_entry.cvterm_id,
                                                      dbxref_id=corresponding_dbxref_entry.dbxref_id,
                                                      is_for_definition=0)
                self.add_and_flush(cvterm_dbxref_entry)
                self.printer.print("Inserted cross reference '" + crossref + "' for CV term '"
                                   + cvterm_entry.name + "'")
                self._crossref_inserts += 1
                edited_entries.append(cvterm_dbxref_entry)

        # Delete cross references not present in file
        for cvterm_dbxref_entry in existing_crossrefs:                              # type: cv.CvTermDbxRef
            matching_crossrefs = utils.filter_objects(edited_entries, dbxref_id=cvterm_dbxref_entry.dbxref_id)
            if not matching_crossrefs:

                # Delete cross reference
                self.session.delete(cvterm_dbxref_entry)
                self.printer.print("Deleted cross reference with dbxref-ID " + str(cvterm_dbxref_entry.dbxref_id)
                                   + " for CV term '" + cvterm_entry.name + "'")
                self._crossref_deletes += 1

        return edited_entries

    def _handle_relationships(self, term: pronto.Term, subject_cvterm_entry: cv.CvTerm,
                              all_cvterm_entries: Dict[str, cv.CvTerm]) -> List[cv.CvTermRelationship]:
        """Inserts, updates or deletes relationships between CV terms"""

        # Get existing relationships for this subject in the database
        subject_db_authority = split_dbxref(term.id)[0]
        existing_relationships = self.query_table(cv.CvTermRelationship,
                                                  subject_id=subject_cvterm_entry.cvterm_id).all()
        edited_entries = []                                                         # type: List[cv.CvTermRelationship]

        # Loop over all relationships for this subject in the file
        for relationship, object_terms in term.relations.items():          # type: pronto.Relationship, pronto.TermList

            # Get relationship type and corresponding term from database
            relationship_name = relationship.obo_name
            if relationship_name not in self._relationship_terms:
                if relationship_name != "can_be":
                    self.printer.print("WARNING: Relationship term '" + relationship_name + "' not present in database")
                continue
            type_cvterm = self._relationship_terms[relationship_name]                # type: cv.CvTerm

            # Loop over all objects
            for object_term in object_terms:                                        # type: pronto.Term

                # Get corresponding object CV term from database
                if split_dbxref(object_term.id)[0] != subject_db_authority:
                    continue
                if object_term.id not in all_cvterm_entries:
                    self.printer.print("WARNING: CV term for dbxref '" + object_term.id + "' not present in database")
                    continue
                object_cvterm_entry = all_cvterm_entries[object_term.id]            # type: cv.CvTerm

                # Check if the relationship already exists in the database
                matching_relationships = utils.filter_objects(existing_relationships, type_id=type_cvterm.cvterm_id,
                                                              object_id=object_cvterm_entry.cvterm_id)
                if matching_relationships:

                    # Save the existing relationship
                    edited_entries.append(matching_relationships[0])
                else:

                    # Insert relationship
                    cvterm_relationship_entry = cv.CvTermRelationship(subject_id=subject_cvterm_entry.cvterm_id,
                                                                      object_id=object_cvterm_entry.cvterm_id,
                                                                      type_id=type_cvterm.cvterm_id)
                    self.add_and_flush(cvterm_relationship_entry)
                    self.printer.print("Inserted relationship: '" + subject_cvterm_entry.name + "', '"
                                       + relationship_name + "', '" + object_cvterm_entry.name + "'")
                    self._relationship_inserts += 1
                    edited_entries.append(cvterm_relationship_entry)

        # Delete relationships not present in file
        for cvterm_relationship_entry in existing_relationships:                    # type: cv.CvTermRelationship
            matching_relationships = utils.filter_objects(edited_entries, type_id=cvterm_relationship_entry.type_id,
                                                          object_id=cvterm_relationship_entry.object_id)
            if not matching_relationships:

                # Delete relationship
                self.session.delete(cvterm_relationship_entry)
                self.printer.print("Deleted relationship for CV term '" + subject_cvterm_entry.name + "'")
                self._relationship_deletes += 1

        return edited_entries

    def _mark_obsolete_terms(self, ontology_terms: Dict[str, pronto.Term], default_db: general.Db) -> List[cv.CvTerm]:
        """Marks obsolete CV terms in the database"""

        # Loop over all dbxref entries in the database
        existing_entries = self.query_table(general.DbxRef, db_id=default_db.db_id).all()
        marked_entries = []
        for dbxref_entry in existing_entries:                                       # type: general.DbxRef

            # Check if the dbxref is also present in the input file
            dbxref = create_dbxref(default_db.name, dbxref_entry.accession, dbxref_entry.version)
            if dbxref not in ontology_terms:

                # Find the corresponding CV term in the database
                cvterm_entry = self.query_table(cv.CvTerm, dbxref_id=dbxref_entry.dbxref_id,
                                                is_relationshiptype=0).first()      # type: cv.CvTerm
                if not cvterm_entry:
                    continue

                # Mark the CV term as obsolete, if necessary
                if self._mark_cvterm_as_obsolete(cvterm_entry):
                    self.printer.print("Marked CV term '" + cvterm_entry.name + "' as obsolete")
                    marked_entries.append(cvterm_entry)
                    self._cvterm_deletes += 1

        return marked_entries

    def _print_statistics(self):
        """Prints a summary of the changes applied to the database"""
        print("Successfully imported an ontology into the database.")
        print("DBs: " + str(self._db_inserts) + " insertions")
        print("CVs: " + str(self._cv_inserts) + " insertions")
        print("dbxrefs: " + str(self._dbxref_inserts) + " insertions, " + str(self._dbxref_updates) + " updates")
        print("CV terms: " + str(self._cvterm_inserts) + " insertions, " + str(self._cvterm_updates)
              + " updates, " + str(self._cvterm_deletes) + " marked as obsolete")
        print("comments: " + str(self._comment_inserts) + " insertions, " + str(self._comment_updates)
              + " updates, " + str(self._comment_deletes) + " deletions")
        print("synonyms: " + str(self._synonym_inserts) + " insertions, " + str(self._synonym_updates)
              + " updates, " + str(self._synonym_deletes) + " deletions")
        print("cross references: " + str(self._crossref_inserts) + " insertions, " + str(self._crossref_updates)
              + " updates, " + str(self._crossref_deletes) + " deletions")
        print("relationships: " + str(self._relationship_inserts) + " insertions, " + str(self._relationship_updates)
              + " updates, " + str(self._relationship_deletes) + " deletions")

    def _initiate_global_arrays(self, default_db_id: int, comment_id: int) -> (
            List[cv.Cv], List[general.DbxRef], List[cv.CvTerm], List[cv.CvTermProp], List[cv.CvTermSynonym],
            List[cv.CvTermDbxRef], List[cv.CvTermRelationship]):
        """Initiates global arrays that can be used to minimize the number of queries to the database"""

        all_cvs = self.query_table(cv.Cv).all()                                    # type: List[cv.Cv]
        all_dbxrefs = self.query_table(general.DbxRef, db_id=default_db_id).all()  # type: List[general.DbxRef]
        all_cvterms = self.query_table(cv.CvTerm).all()                            # type: List[cv.CvTerm]
        all_comments = self.query_table(cv.CvTermProp, type_id=comment_id).all()   # type: List[cv.CvTermProp]
        all_synonyms = self.query_table(cv.CvTermSynonym).all()                    # type: List[cv.CvTermSynonym]
        all_crossrefs = self.query_table(cv.CvTermDbxRef).all()                    # type: List[cv.CvTermDbxRef]
        all_relationships = self.query_table(cv.CvTermRelationship).all()          # type: List[cv.CvTermRelationship]

        return all_cvs, all_dbxrefs, all_cvterms, all_comments, all_synonyms, all_crossrefs, all_relationships


def parse_ontology(filename: str, file_format="obo") -> pronto.Ontology:
    """Function parsing an OBO/OWL file"""
    if file_format == "owl":
        return pronto.Ontology(filename, parser="OwlXMLParser")
    else:
        return pronto.Ontology(filename, parser="OboParser")


def get_default_namespace(ontology: pronto.Ontology) -> str:
    """Retrieves the default namespace of a given ontology"""
    default_namespace = ""
    if "default-namespace" in ontology.meta:
        default_namespace = ontology.meta["default-namespace"][0]
    return default_namespace


def filter_ontology_by_db(ontology: pronto.Ontology, db_authority: str) -> Dict[str, pronto.Term]:
    """Filters the terms in an ontology, only retaining those stemming from a given database"""
    filtered_terms = {}                                                                 # type: Dict[str, pronto.Term]
    for term in ontology:                                                               # type: pronto.Term
        db = split_dbxref(term.id)[0]
        if db == db_authority:
            filtered_terms[term.id] = term
    return filtered_terms


def extract_comment(term: pronto.Term) -> str:
    """Extracts comments from an ontology term"""
    comment = ""
    if "comment" in term.other:
        comment = term.other["comment"][0]
    return comment


def extract_synonyms(term: pronto.Term) -> List[str]:
    """Extracts synonyms from an ontology term"""
    synonyms = []
    for synonym in term.synonyms:
        synonyms.append(synonym.desc)
    return synonyms


def extract_cross_references(term: pronto.Term) -> List[str]:
    """Extracts cross references from an ontology term"""
    cross_references = []
    if "alt_id" in term.other:
        cross_references.extend(term.other["alt_id"])
    if "xref" in term.other:
        for xref in term.other["xref"]:                                                 # type: str
            cross_references.append(xref.split(" ")[0])
    return cross_references


def create_cvterm_entry(term: pronto.Term, cv_id: int, dbxref_id: int) -> cv.CvTerm:
    """Creates a CV term entry for the database from an ontology term"""
    is_obsolete = False
    if "is_obsolete" in term.other and term.other["is_obsolete"][0].lower() == "true":
        is_obsolete = True
    description = None
    if term.desc:
        description = term.desc
    cvterm_entry = cv.CvTerm(cv_id=cv_id, dbxref_id=dbxref_id, name=term.name, definition=description,
                             is_obsolete=int(is_obsolete))
    return cvterm_entry


def update_cvterm_properties(existing_cvterm_entry: cv.CvTerm, new_cvterm_entry: cv.CvTerm) -> bool:
    """Updates the properties of a CV term in the database"""
    updated = False
    if new_cvterm_entry.cv_id != existing_cvterm_entry.cv_id:
        existing_cvterm_entry.cv_id = new_cvterm_entry.cv_id
        updated = True
    if new_cvterm_entry.name != existing_cvterm_entry.name:
        existing_cvterm_entry.name = new_cvterm_entry.name
        updated = True
    if new_cvterm_entry.definition != existing_cvterm_entry.definition:
        existing_cvterm_entry.definition = new_cvterm_entry.definition
        updated = True
    if new_cvterm_entry.is_obsolete != existing_cvterm_entry.is_obsolete:
        existing_cvterm_entry.is_obsolete = new_cvterm_entry.is_obsolete
        updated = True
    return updated


def split_dbxref(dbxref: str) -> (str, str, str):
    """Splits a database cross reference string in the format db:accession:version into its constituents"""
    split_id = dbxref.split(":")
    if len(split_id) < 2 or len(split_id) > 3:
        raise AttributeError("dbxref must consist of 2 or 3 elements, separated by semicolon")
    db = split_id[0]
    accession = split_id[1]
    version = ""
    if len(split_id) == 3:
        version = split_id[2]
    return db, accession, version


def create_dbxref(db: str, accession: str, version="") -> str:
    """Creates a database cross reference string in the format db:accession:version from the constituents"""
    if not db or not accession:
        raise AttributeError("db and accession must be given for a dbxref")
    dbxref = db + ":" + accession
    if version:
        dbxref = dbxref + ":" + version
    return dbxref



from typing import List, Set, Dict, Union
from Bio.UniProt import GOA
from . import iobase, ontology
from .. import utils
from ..orm import general, cv, organism, pub, sequence


class GAFClient(iobase.ChadoClient):

    def __init__(self, uri: str, verbose=False, test_environment=False):
        """Constructor"""

        # Connect to database
        self.test_environment = test_environment
        if self.test_environment:
            self.printer = utils.VerbosePrinter(verbose)
        else:
            super().__init__(uri, verbose)

        # Load essentials
        if not self.test_environment:
            self._load_essentials()

    def __del__(self):
        """Destructor - disconnect from database"""
        if not self.test_environment:
            super().__del__()

    def _load_essentials(self) -> None:
        """Loads essential database entries"""
        self._part_of_term = self._load_cvterm_from_cv("part_of", "relationship")
        self._derives_from_term = self._load_cvterm_from_cv("derives_from", "sequence")
        self._date_term = self._load_cvterm("date")
        self._evidence_term = self._load_cvterm("evidence")
        self._synonym_term = self._load_cvterm("synonym")
        self._taxon_term = self._load_cvterm("taxonId")
        self._default_pub = self._load_pub("null")
        self._go_db = self._load_db("GO")
        self._product_db = self._load_db("PRODUCT")

    def _extract_feature_type(self, feature_entry: sequence.Feature) -> str:
        """Extracts the type of a feature by a database query"""
        cvterm_entry = self.query_first(cv.CvTerm, cvterm_id=feature_entry.type_id)
        return cvterm_entry.name

    def _is_featuretype_valid(self, feature_entry: sequence.Feature) -> bool:
        """Checks if a feature has a valid type"""
        featuretype = self._extract_feature_type(feature_entry)
        valid_featuretypes = self._gene_types() + self._transcript_types() + self._protein_types()
        if featuretype.lower() not in valid_featuretypes:
            return False
        return True

    def _extract_requested_feature(self, feature_entry: sequence.Feature, annotation_level: str
                                   ) -> sequence.Feature:
        """Extracts a gene, transcript or protein entry from the database, depending on the desired annotation level"""
        if annotation_level == "gene":
            requested_entry = self._extract_gene_of_feature(feature_entry)
        elif annotation_level == "transcript":
            requested_entry = self._extract_transcript_of_feature(feature_entry)
        elif annotation_level == "protein":
            requested_entry = self._extract_polypeptide_of_feature(feature_entry)
        else:
            requested_entry = feature_entry
        if not requested_entry:
            self.printer.print("WARNING: No " + annotation_level + " associated with feature '"
                               + feature_entry.uniquename + "' available in the database.")
            requested_entry = feature_entry
        return requested_entry

    def _extract_gene_of_feature(self, feature_entry: sequence.Feature) -> Union[None, sequence.Feature]:
        """Extracts the gene entry for a given feature, which can be a gene, transcript or polypeptide"""
        featuretype = self._extract_feature_type(feature_entry)
        if featuretype.lower() in self._gene_types():
            gene_entry = feature_entry
        elif featuretype.lower() in self._transcript_types():
            gene_entry = self.query_parent_features(feature_entry.feature_id, [self._part_of_term.cvterm_id]).first()
        elif featuretype.lower() in self._protein_types():
            transcript_entry = self.query_parent_features(
                feature_entry.feature_id, [self._derives_from_term.cvterm_id]).first()
            if transcript_entry:
                gene_entry = self.query_parent_features(
                    transcript_entry.feature_id, [self._part_of_term.cvterm_id]).first()
            else:
                gene_entry = None
        else:
            gene_entry = None
        return gene_entry

    def _extract_transcript_of_feature(self, feature_entry: sequence.Feature) -> Union[None, sequence.Feature]:
        """Extracts the transcript entry for a given feature, which can be a gene, transcript or polypeptide"""
        featuretype = self._extract_feature_type(feature_entry)
        if featuretype.lower() in self._gene_types():
            transcript_entry = self.query_child_features(feature_entry.feature_id, self._part_of_term.cvterm_id).first()
        elif featuretype.lower() in self._transcript_types():
            transcript_entry = feature_entry
        elif featuretype.lower() in self._protein_types():
            transcript_entry = self.query_parent_features(
                feature_entry.feature_id, [self._derives_from_term.cvterm_id]).first()
        else:
            transcript_entry = None
        return transcript_entry

    def _extract_polypeptide_of_feature(self, feature_entry: sequence.Feature) -> Union[None, sequence.Feature]:
        """Extracts the polypeptide entry for a given feature, which can be a gene, transcript or polypeptide"""
        featuretype = self._extract_feature_type(feature_entry)
        if featuretype.lower() in self._gene_types():
            transcript_entry = self.query_child_features(feature_entry.feature_id, self._part_of_term.cvterm_id).first()
            if transcript_entry:
                polypeptide_entry = self.query_child_features(
                    transcript_entry.feature_id, self._derives_from_term.cvterm_id).first()
            else:
                polypeptide_entry = None
        elif featuretype.lower() in self._transcript_types():
            polypeptide_entry = self.query_child_features(
                feature_entry.feature_id, self._derives_from_term.cvterm_id).first()
        elif featuretype.lower() in self._protein_types():
            polypeptide_entry = feature_entry
        else:
            polypeptide_entry = None
        return polypeptide_entry

    def _convert_evidence_code(self, abbreviation: str) -> str:
        """Converts the abbreviation for an evidence code into the spelled-out version, if applicable"""
        if abbreviation in self._abbreviations_to_evidence_codes():
            return self._abbreviations_to_evidence_codes()[abbreviation]
        return ""

    def _back_convert_evidence_code(self, evidence_code: str) -> str:
        """Converts a spelled-out evidence code into its abbreviation, if applicable"""
        if evidence_code.lower() in self._evidence_codes_to_abbreviations():
            return self._evidence_codes_to_abbreviations()[evidence_code.lower()]
        if evidence_code.upper() in self._abbreviations_to_evidence_codes():
            return evidence_code.upper()
        return ""

    def _convert_namespace(self, namespace: str) -> str:
        """Converts a GO namespace into its abbreviation, if applicable"""
        if namespace in self._namespace_to_abbreviation():
            return self._namespace_to_abbreviation()[namespace]
        return ""

    @staticmethod
    def _abbreviations_to_evidence_codes() -> Dict[str, str]:
        """Lists the GO evidence codes and their respective abbreviations"""
        return {
            "EXP": "Inferred from Experiment",
            "IDA": "Inferred from Direct Assay",
            "IPI": "Inferred from Physical Interaction",
            "IMP": "Inferred from Mutant Phenotype",
            "IGI": "Inferred from Genetic Interaction",
            "IEP": "Inferred from Expression Pattern",
            "HTP": "Inferred from High Throughput Experiment",
            "HDA": "Inferred from High Throughput Direct Assay",
            "HMP": "Inferred from High Throughput Mutant Phenotype",
            "HGI": "Inferred from High Throughput Genetic Interaction",
            "HEP": "Inferred from High Throughput Expression Pattern",
            "ISS": "Inferred from Sequence or structural Similarity",
            "ISO": "Inferred from Sequence Orthology",
            "ISA": "Inferred from Sequence Alignment",
            "ISM": "Inferred from Sequence Model",
            "IGC": "Inferred from Genomic Context",
            "IBA": "Inferred from Biological aspect of Ancestor",
            "IBD": "Inferred from Biological aspect of Descendant",
            "IKR": "Inferred from Key Residues",
            "IRD": "Inferred from Rapid Divergence",
            "RCA": "Inferred from Reviewed Computational Analysis",
            "TAS": "Traceable Author Statement",
            "NAS": "Non-traceable Author Statement",
            "IC": "Inferred by Curator",
            "ND": "No biological Data available",
            "IEA": "Inferred from Electronic Annotation",
            "NR": "Not recorded"
        }

    @staticmethod
    def _default_evidence_code() -> str:
        """Yields the default GO evidence code"""
        return "NR"

    def _evidence_codes_to_abbreviations(self) -> Dict[str, str]:
        """Lists the GO evidence codes and their respective abbreviations"""
        return {
            evidence_code.lower(): abbreviation
            for abbreviation, evidence_code
            in self._abbreviations_to_evidence_codes().items()
        }

    @staticmethod
    def _namespace_to_abbreviation() -> Dict[str, str]:
        """Lists the GO namespaces and their respective abbreviations"""
        return {
            "cellular_component": "C",
            "molecular_function": "F",
            "biological_process": "P"
        }

    @staticmethod
    def _db_reference_identifiers() -> List[str]:
        """Lists the permitted database authorities for the GAF DB:Reference column"""
        return ["PMID", "GO_REF", "SGD_REF"]

    @staticmethod
    def _default_db_reference() -> str:
        """Yields the default value for the GAF DB:Reference column"""
        return "GO_REF:0000002"

    @staticmethod
    def _gene_types() -> List[str]:
        """Lists considered gene types"""
        return ["gene", "pseudogene"]

    @staticmethod
    def _transcript_types() -> List[str]:
        """Lists considered transcript types"""
        return ["mrna", "rrna", "trna", "snrna", "ncrna", "scrna", "snorna", "pseudogenic_transcript"]

    @staticmethod
    def _protein_types() -> List[str]:
        """Lists considered protein types"""
        return ["polypeptide"]


class GAFImportClient(GAFClient):
    """Class for importing genomic data from GAF files into Chado"""

    def load(self, filename: str, organism_name: str, annotation_level: str):
        """Import data from a GAF file into a Chado database"""

        # Load dependencies
        default_organism = self._load_organism(organism_name)
        features_with_product = set()

        # Loop over all records in the GAF file
        with open(filename) as f:
            for gaf_record in GOA.gafiterator(f):

                # Import this record into the database
                self._load_gaf_record(gaf_record, default_organism, annotation_level, features_with_product)

        # Commit changes
        self.session.commit()

    def _load_gaf_record(self, gaf_record: dict, organism_entry: organism.Organism, annotation_level: str,
                         features_with_product: Set[str]) -> None:
        """Imports data from a GAF record into a Chado database"""

        # Get the feature entries
        gaf_feature_entry = self._load_feature(gaf_record, organism_entry)
        if not gaf_feature_entry:
            return
        requested_feature_entry = self._extract_requested_feature(gaf_feature_entry, annotation_level)
        gene_feature_entry = self._extract_gene_of_feature(requested_feature_entry)

        # Update/insert gene name and synonym(s)
        self._handle_name(gaf_record, gene_feature_entry)
        self._handle_synonyms(gaf_record, gene_feature_entry)

        # Update/insert a feature_cvterm entry for the gene product
        if requested_feature_entry.uniquename not in features_with_product:
            self._handle_product_term(gaf_record, requested_feature_entry)
            features_with_product.add(requested_feature_entry.uniquename)

        # Update/insert a feature_cvterm entry for the ontology term
        feature_cvterm_entry = self._handle_ontology_term(gaf_record, requested_feature_entry)
        if not feature_cvterm_entry:
            return

        # Update/insert feature_cvtermprop entries for date and evidence code
        self._handle_properties(gaf_record, feature_cvterm_entry, requested_feature_entry)

        # Update/insert feature_cvterm_dbxref entries
        self._handle_crossrefs(gaf_record, feature_cvterm_entry, requested_feature_entry)

        # Update/insert feature_cvterm_pub entries
        self._handle_publications(gaf_record, feature_cvterm_entry, requested_feature_entry)

    def _load_feature(self, gaf_record: dict, organism_entry: organism.Organism) -> Union[None, sequence.Feature]:
        """Loads a feature entry from the database"""
        feature_name = gaf_record["DB_Object_ID"].strip()
        if not feature_name:
            return None
        feature_entry = self.query_first(sequence.Feature, organism_id=organism_entry.organism_id,
                                         uniquename=feature_name)
        return feature_entry

    def _handle_name(self, gaf_record: dict, feature_entry: sequence.Feature) -> None:
        """Inserts or updates the name of a feature"""
        name = gaf_record["DB_Object_Symbol"].strip()
        if name and name != feature_entry.uniquename and name != feature_entry.name:
            feature_entry.name = name
            self.printer.print("Updated name '" + feature_entry.name + "' for feature '"
                               + feature_entry.uniquename + "'")

    def _handle_synonyms(self, gaf_record: dict, feature_entry: sequence.Feature) -> List[sequence.FeatureSynonym]:
        """Inserts or updates entries in the 'synonym' and 'feature_synonym' tables and returns the latter"""

        # Extract existing synonyms for this feature from the database
        existing_feature_synonyms = self.query_feature_synonym_by_type(
            feature_entry.feature_id, [self._synonym_term.cvterm_id]).all()
        all_feature_synonyms = []

        # Loop over all synonyms for this feature in the GFF record
        synonyms = gaf_record["Synonym"]
        for synonym in synonyms:

            # Insert/update entry in the 'synonym' table
            if not synonym.strip():
                continue
            new_synonym_entry = sequence.Synonym(name=synonym, type_id=self._synonym_term.cvterm_id,
                                                 synonym_sgml=synonym)
            synonym_entry = self._handle_synonym(new_synonym_entry)

            # Insert/update entry in the 'feature_synonym' table
            new_feature_synonym_entry = sequence.FeatureSynonym(
                synonym_id=synonym_entry.synonym_id, feature_id=feature_entry.feature_id,
                pub_id=self._default_pub.pub_id)
            feature_synonym_entry = self._handle_feature_synonym(new_feature_synonym_entry, existing_feature_synonyms,
                                                                 synonym, feature_entry.uniquename)
            all_feature_synonyms.append(feature_synonym_entry)

        return all_feature_synonyms

    def _handle_ontology_term(self, gaf_record: dict, feature_entry: sequence.Feature
                              ) -> Union[None, sequence.FeatureCvTerm]:
        """Inserts or updates an entry in the 'feature_cvterm' table and returns it"""

        # Extract ontology identifier from the GAF record and split it into db, accession, version
        ontology_term = gaf_record["GO_ID"].strip()
        (db_authority, accession, version) = ontology.split_dbxref(ontology_term)
        publication = self._extract_primary_publication(gaf_record)

        # Get entry from 'db' table
        db_entry = self.query_first(general.Db, name=db_authority)
        if not db_entry:
            self.printer.print("WARNING: Ontology '" + db_authority + "' not present in database.")
            return None

        # Get entry from 'dbxref' table
        dbxref_entry = self.query_first(general.DbxRef, db_id=db_entry.db_id, accession=accession)
        if not dbxref_entry:
            self.printer.print("WARNING: Ontology term '" + ontology_term + "' not present in database.")
            return None

        # Get entry from 'cvterm' table
        cvterm_entry = self.query_first(cv.CvTerm, dbxref_id=dbxref_entry.dbxref_id)
        if not cvterm_entry:
            self.printer.print("WARNING: CV term for ontology term '" + ontology_term
                               + "' not present in database.")
            return None

        # Insert/update entry in the 'pub' table
        if publication:
            new_pub_entry = pub.Pub(uniquename=publication, type_id=self._default_pub.type_id)
            pub_entry = self._handle_pub(new_pub_entry)
        else:
            pub_entry = self._default_pub

        # Extract existing ontology terms for this feature from the database
        existing_feature_cvterms = self.query_feature_cvterm_by_ontology(
            feature_entry.feature_id, db_entry.db_id).all()

        # Insert/update entry in the 'feature_cvterm' table
        new_feature_cvterm_entry = sequence.FeatureCvTerm(feature_id=feature_entry.feature_id,
                                                          cvterm_id=cvterm_entry.cvterm_id,
                                                          pub_id=pub_entry.pub_id,
                                                          is_not=("NOT" in gaf_record["Qualifier"]))
        feature_cvterm_entry = self._handle_feature_cvterm(new_feature_cvterm_entry, existing_feature_cvterms,
                                                           cvterm_entry.name, feature_entry.uniquename)
        return feature_cvterm_entry

    def _handle_product_term(self, gaf_record: dict, feature_entry: sequence.Feature
                             ) -> Union[None, sequence.FeatureCvTerm]:
        """Inserts or updates an entry in the 'feature_cvterm' table and returns it"""

        # Extract the product name from the GAF record
        product = gaf_record["DB_Object_Name"].strip()
        if not product:
            return None
        publication = self._extract_primary_publication(gaf_record)

        # Insert/update entry in the 'db' table
        new_db_entry = general.Db(name="PRODUCT")
        db_entry = self._handle_db(new_db_entry)

        # Insert/update entry in the 'dbxref' table
        new_dbxref_entry = general.DbxRef(db_id=db_entry.db_id, accession=product)
        dbxref_entry = self._handle_dbxref(new_dbxref_entry, db_entry.name)

        # Insert/update entry in the 'cv' table
        new_cv_entry = cv.Cv(name="genedb_products")
        cv_entry = self._handle_cv(new_cv_entry)

        # Insert/update entry in the 'cvterm' table
        new_cvterm_entry = cv.CvTerm(cv_id=cv_entry.cv_id, dbxref_id=dbxref_entry.dbxref_id, name=product)
        cvterm_entry = self._handle_cvterm(new_cvterm_entry, cv_entry.name)

        # Insert/update entry in the 'pub' table
        if publication:
            new_pub_entry = pub.Pub(uniquename=publication, type_id=self._default_pub.type_id)
            pub_entry = self._handle_pub(new_pub_entry)
        else:
            pub_entry = self._default_pub

        # Extract existing product terms for this feature from the database
        existing_feature_cvterms = self.query_feature_cvterm_by_ontology(
            feature_entry.feature_id, db_entry.db_id).all()

        # Insert/update entry in the 'feature_cvterm' table
        new_feature_cvterm_entry = sequence.FeatureCvTerm(feature_id=feature_entry.feature_id,
                                                          cvterm_id=cvterm_entry.cvterm_id,
                                                          pub_id=pub_entry.pub_id)
        feature_cvterm_entry = self._handle_feature_cvterm(new_feature_cvterm_entry, existing_feature_cvterms,
                                                           cvterm_entry.name, feature_entry.uniquename)
        return feature_cvterm_entry

    def _handle_properties(self, gaf_record: dict, feature_cvterm_entry: sequence.FeatureCvTerm,
                           feature_entry: sequence.Feature) -> List[sequence.FeatureCvTermProp]:
        """Inserts or updates entries in the 'feature_cvtermprop' table and returns them"""

        # Extract existing properties for this feature_cvterm from the database
        existing_feature_cvtermprops = self.query_all(sequence.FeatureCvTermProp,
                                                      feature_cvterm_id=feature_cvterm_entry.feature_cvterm_id)

        # Insert/update entry for 'date'
        date = gaf_record["Date"].strip()
        new_feature_cvtermprop_entry = sequence.FeatureCvTermProp(
            feature_cvterm_id=feature_cvterm_entry.feature_cvterm_id, type_id=self._date_term.cvterm_id, value=date)
        date_feature_cvtermprop_entry = self._handle_feature_cvtermprop(
            new_feature_cvtermprop_entry, existing_feature_cvtermprops, "date", feature_entry.uniquename)

        # Insert/update entry for 'evidence'
        evidence_code = self._convert_evidence_code(gaf_record["Evidence"])
        new_feature_cvtermprop_entry = sequence.FeatureCvTermProp(
            feature_cvterm_id=feature_cvterm_entry.feature_cvterm_id, type_id=self._evidence_term.cvterm_id,
            value=evidence_code)
        evidence_feature_cvtermprop_entry = self._handle_feature_cvtermprop(
            new_feature_cvtermprop_entry, existing_feature_cvtermprops, "evidence", feature_entry.uniquename)

        return [date_feature_cvtermprop_entry, evidence_feature_cvtermprop_entry]

    def _handle_crossrefs(self, gaf_record: dict, feature_cvterm_entry: sequence.FeatureCvTerm,
                          feature_entry: sequence.Feature) -> List[sequence.FeatureCvTermDbxRef]:
        """Inserts or updates an entry in the 'feature_cvterm_dbxref' table and returns it"""

        # Extract existing cross references for this feature_cvterm from the database
        existing_feature_cvterm_dbxrefs = self.query_all(sequence.FeatureCvTermDbxRef,
                                                         feature_cvterm_id=feature_cvterm_entry.feature_cvterm_id)
        all_feature_cvterm_dbxrefs = []

        # Loop over all cross references of the given GAF record
        crossrefs = self._extract_cross_references(gaf_record)
        for crossref in crossrefs:

            # Split database cross reference (dbxref) into db, accession, version
            (db_authority, accession, version) = ontology.split_dbxref(crossref)

            # Insert/update entry in the 'db' table
            new_db_entry = general.Db(name=db_authority)
            db_entry = self._handle_db(new_db_entry)

            # Insert/update entry in the 'dbxref' table
            new_dbxref_entry = general.DbxRef(db_id=db_entry.db_id, accession=accession, version=version)
            dbxref_entry = self._handle_dbxref(new_dbxref_entry, db_authority)

            # Insert/update entry in the 'feature_cvterm_dbxref' table
            new_feature_cvterm_dbxref_entry = sequence.FeatureCvTermDbxRef(
                feature_cvterm_id=feature_cvterm_entry.feature_cvterm_id, dbxref_id=dbxref_entry.dbxref_id)
            feature_cvterm_dbxref_entry = self._handle_feature_cvterm_dbxref(
                new_feature_cvterm_dbxref_entry, existing_feature_cvterm_dbxrefs, crossref, feature_entry.uniquename)
            all_feature_cvterm_dbxrefs.append(feature_cvterm_dbxref_entry)

        return all_feature_cvterm_dbxrefs

    def _handle_publications(self, gaf_record: dict, feature_cvterm_entry: sequence.FeatureCvTerm,
                             feature_entry: sequence.Feature) -> List[sequence.FeatureCvTermPub]:
        """Inserts or updates entries in the 'feature_cvterm_pub' table and returns them"""

        # Extract existing publications for this feature_cvterm from the database
        existing_feature_cvterm_pubs = self.query_all(sequence.FeatureCvTermPub,
                                                      feature_cvterm_id=feature_cvterm_entry.feature_cvterm_id)
        all_feature_cvterm_pubs = []

        # Loop over all publications of the given GAF record
        publications = self._extract_secondary_publications(gaf_record)
        for publication in publications:

            # Insert/update entry in the 'pub' table
            new_pub_entry = pub.Pub(uniquename=publication, type_id=self._default_pub.type_id)
            pub_entry = self._handle_pub(new_pub_entry)

            # Insert/update entry in the 'feature_cvterm_pub' table
            new_feature_cvterm_pub_entry = sequence.FeatureCvTermPub(
                feature_cvterm_id=feature_cvterm_entry.feature_cvterm_id, pub_id=pub_entry.pub_id)
            feature_cvterm_pub_entry = self._handle_feature_cvterm_pub(
                new_feature_cvterm_pub_entry, existing_feature_cvterm_pubs, publication, feature_entry.uniquename)
            all_feature_cvterm_pubs.append(feature_cvterm_pub_entry)

        return all_feature_cvterm_pubs

    def _extract_cross_references(self, gaf_record: dict) -> List[str]:
        """Extracts the database cross references of a GAF record"""
        potential_crossrefs = gaf_record["With"] + gaf_record["DB:Reference"]
        crossrefs = []
        pubs = [self._extract_primary_publication(gaf_record)] + self._extract_secondary_publications(gaf_record)
        for crossref in potential_crossrefs:
            if crossref.strip() and crossref not in pubs:
                crossrefs.append(crossref)
        return crossrefs

    @staticmethod
    def _extract_primary_publication(gaf_record: dict) -> Union[None, str]:
        """Extracts the primary publication of a GAF record"""
        potential_publications = gaf_record["DB:Reference"]
        for publication in potential_publications:
            if ":" not in publication:
                return publication.strip()
            (db_authority, accession, version) = ontology.split_dbxref(publication)
            if db_authority == "PMID":
                return publication.strip()
        return None

    @staticmethod
    def _extract_secondary_publications(gaf_record: dict) -> List[str]:
        """Extracts the secondary publications of a GAF record"""
        potential_publications = gaf_record["DB:Reference"]
        secondary_publications = []
        for publication in potential_publications:
            if ":" not in publication:
                secondary_publications.append(publication.strip())
                continue
            (db_authority, accession, version) = ontology.split_dbxref(publication)
            if db_authority == "PMID":
                secondary_publications.append(publication.strip())
        if secondary_publications:
            return secondary_publications[1:]
        else:
            return secondary_publications


class GAFExportClient(GAFClient):
    """Class for exporting gene annotation data from Chado to GAF files"""

    def export(self, gaf_filename: str, organism_name: str, database_authority: str, annotation_level: str,
               include_obsolete_features=False) -> None:

        # Load dependencies
        organism_entry = self._load_organism(organism_name)
        taxon_id = self._extract_taxon_id(organism_entry)

        # Open GAF file and write header
        with utils.open_file_write(gaf_filename) as gaf_handle:
            self._write_gaf_header(gaf_handle)

            # Get all feature_cvterms associated with GO terms
            go_feature_cvterm_entries = self.query_feature_cvterm_by_ontology_and_organism(
                organism_entry.organism_id, self._go_db.db_id).order_by(
                sequence.Feature.uniquename, general.DbxRef.accession).all()

            # Loop over all feature_cvterms
            for go_feature_cvterm_entry in go_feature_cvterm_entries:

                # Create a GFF record for this feature_cvterm
                go_feature_entry = self.query_first(sequence.Feature, feature_id=go_feature_cvterm_entry.feature_id)
                if self._is_featuretype_valid(go_feature_entry) \
                        and (include_obsolete_features or not go_feature_entry.is_obsolete):
                    self._export_gaf_record(go_feature_entry, go_feature_cvterm_entry, database_authority, taxon_id,
                                            annotation_level, gaf_handle)

        # Information
        self.printer.print("Exported GAF data for organism " + organism_name + " to " + gaf_filename + ".")

    @staticmethod
    def _write_gaf_header(file_handle):
        """Prints the header of a GAF file"""
        file_handle.write("!gaf-version: 1.0\n")

    def _export_gaf_record(self, feature_entry: sequence.Feature, feature_cvterm_entry: sequence.FeatureCvTerm,
                           database_authority: str, taxon_id: str, annotation_level: str, file_handle):
        """Exports a GAF record for a given feature and GO term"""

        try:
            # Create GAF record
            requested_feature = self._extract_requested_feature(feature_entry, annotation_level)
            gene_feature = self._extract_gene_of_feature(requested_feature)
            gaf_record = self._create_gaf_record(feature_cvterm_entry, requested_feature, database_authority, taxon_id)

            # Query various tables to gather information related to the feature_cvterm
            go_id = self._extract_feature_cvterm_ontology_term(feature_cvterm_entry)
            go_namespace = self._extract_go_namespace(feature_cvterm_entry)
            feature_cvterm_properties = self._extract_feature_cvterm_properties(feature_cvterm_entry)
            feature_cvterm_dbxrefs = self._extract_feature_cvterm_dbxrefs(feature_cvterm_entry)
            feature_cvterm_publications = self._extract_feature_cvterm_publications(feature_cvterm_entry)
            featuretype = self._extract_feature_type(requested_feature)
            gene_name = self._extract_feature_name(gene_feature)
            gene_synonyms = self._extract_feature_synonyms(gene_feature)
            product_name = self._extract_product_name(feature_entry)

            # Add attributes to the GAF record
            self._add_gaf_go_id(gaf_record, go_id)
            self._add_gaf_aspect(gaf_record, go_namespace)
            self._add_gaf_annotation_date(gaf_record, feature_cvterm_properties)
            self._add_gaf_evidence_code(gaf_record, feature_cvterm_properties)
            self._add_gaf_withfrom_info(gaf_record, feature_cvterm_dbxrefs)
            self._add_gaf_db_references(gaf_record, feature_cvterm_publications, feature_cvterm_dbxrefs)
            self._add_gaf_object_type(gaf_record, featuretype)
            self._add_gaf_object_symbol(gaf_record, gene_name)
            self._add_gaf_synonyms(gaf_record, gene_synonyms)
            self._add_gaf_object_name(gaf_record, product_name)

            # Write the generated GAF record to file
            self._print_gaf_record(file_handle, gaf_record)

        except iobase.DatabaseError as err:
            self.printer.print(err.args)

    def _print_gaf_record(self, file_handle, gaf_record: dict):
        """Prints a GAF record to file"""
        line = "\t".join([self._stringify_gaf_attribute(value) for value in gaf_record.values()])
        file_handle.write(line + "\n")

    @staticmethod
    def _stringify_gaf_attribute(attribute: Union[str, List[str]]) -> str:
        """Converts a GAF attribute from a list to a string, separated by pipes"""
        if isinstance(attribute, str):
            return attribute
        elif isinstance(attribute, list):
            return utils.list_to_string(attribute, "|")
        else:
            return ""

    @staticmethod
    def _create_gaf_record(feature_cvterm_entry: sequence.FeatureCvTerm, feature_entry: sequence.Feature,
                           database_authority: str, taxon_id: str) -> dict:
        """Creates a GAF record"""
        gaf_record = {key: None for key in GOA.GAF10FIELDS}
        gaf_record["DB"] = database_authority
        gaf_record["Assigned_By"] = database_authority
        gaf_record["DB_Object_ID"] = feature_entry.uniquename
        gaf_record["Taxon_ID"] = taxon_id
        if feature_cvterm_entry.is_not:
            gaf_record["Qualifier"] = "NOT"
        return gaf_record

    def _extract_taxon_id(self, organism_entry: organism.Organism) -> str:
        """Extracts the NCBI taxon ID associated with an organism"""
        taxon_id = "taxon:" + organism_entry.abbreviation
        organismprop_entry = self.query_first(organism.OrganismProp, organism_id=organism_entry.organism_id,
                                              type_id=self._taxon_term.cvterm_id)
        if organismprop_entry:
            taxon_id = "taxon:" + organismprop_entry.value
        return taxon_id

    def _extract_feature_cvterm_ontology_term(self, feature_cvterm_entry: sequence.FeatureCvTerm) -> str:
        """Extracts the ontology term associated with a feature_cvterm by a database query"""
        db_and_accession = self.query_feature_cvterm_ontology_terms(
            feature_cvterm_entry.feature_cvterm_id, self._go_db.db_id).first()
        ontology_term = ""
        if db_and_accession:
            ontology_term = ontology.create_dbxref(db_and_accession[0], db_and_accession[1])
        return ontology_term

    def _extract_feature_cvterm_properties(self, feature_cvterm_entry: sequence.FeatureCvTerm) -> Dict[str, str]:
        """Extracts properties associated with a feature_cvterm by a database query"""
        properties = {}
        for property_type, property_value in self.query_feature_cvterm_properties(
                feature_cvterm_entry.feature_cvterm_id).all():
            properties[property_type] = property_value
        return properties

    def _extract_feature_cvterm_dbxrefs(self, feature_cvterm_entry: sequence.FeatureCvTerm) -> List[str]:
        """Extracts database cross references associated with a feature_cvterm by a database query"""
        cross_references = []
        for db_authority, accession in self.query_feature_cvterm_dbxrefs(feature_cvterm_entry.feature_cvterm_id).all():
            crossref = ontology.create_dbxref(db_authority, accession)
            cross_references.append(crossref)
        return cross_references

    def _extract_feature_cvterm_publications(self, feature_cvterm_entry: sequence.FeatureCvTerm) -> List[str]:
        """Extracts publications associated with a feature_cvterm by a database query"""
        publications = []
        for publication, in self.query_feature_cvterm_pubs(feature_cvterm_entry.feature_cvterm_id).all():
            publications.append(publication)
        for publication, in self.query_feature_cvterm_secondary_pubs(feature_cvterm_entry.feature_cvterm_id).all():
            publications.append(publication)
        return publications

    def _extract_go_namespace(self, feature_cvterm_entry: sequence.FeatureCvTerm) -> str:
        """Extracts the namespace of a feature_cvterm by a database query"""
        return self.query_cvterm_namespace(feature_cvterm_entry.cvterm_id).scalar()

    @staticmethod
    def _extract_feature_name(feature_entry: sequence.Feature) -> str:
        """Extracts the name of a given feature"""
        name = ""
        if feature_entry:
            name = feature_entry.name or feature_entry.uniquename
        return name

    def _extract_feature_synonyms(self, feature_entry: sequence.Feature) -> List[str]:
        """Extracts synonyms of a feature by a database query"""
        synonyms = []
        if feature_entry:
            for synonym_type, synonym_name in self.query_feature_synonyms(feature_entry.feature_id).all():
                synonyms.append(synonym_name)
        return synonyms

    def _extract_product_name(self, feature_entry: sequence.Feature) -> str:
        """Extracts the gene product of a feature by a database query"""
        name = ""
        product_feature_cvterm = self.query_feature_cvterm_by_ontology(
            feature_entry.feature_id, self._product_db.db_id).first()
        if product_feature_cvterm:
            name = self.query_first(cv.CvTerm, cvterm_id=product_feature_cvterm.cvterm_id).name
        return name

    @staticmethod
    def _add_gaf_go_id(gaf_record: dict, go_id: str) -> None:
        """Adds the GO term to a GAF record"""
        if not go_id:
            raise iobase.DatabaseError("Missing ontology term for feature '" + gaf_record["DB_Object_ID"] + "'")
        gaf_record["GO_ID"] = go_id

    def _add_gaf_aspect(self, gaf_record: dict, namespace: str) -> None:
        """Adds the 'aspect' (GO namespace) to a GAF record"""
        aspect = self._convert_namespace(namespace)
        if not aspect:
            raise iobase.DatabaseError("Unrecognized GO namespace: '" + namespace + "' for GO term '"
                                       + gaf_record["GO_ID"] + "'")
        gaf_record["Aspect"] = aspect

    def _add_gaf_annotation_date(self, gaf_record: dict, properties: Dict[str, str]) -> None:
        """Adds the annotation date to a GAF record"""
        if "date" not in properties:
            self.printer.print("Missing annotation date for feature '" + gaf_record["DB_Object_ID"]
                               + "' and GO term '" + gaf_record["GO_ID"] + "'")
            gaf_record["Date"] = utils.current_date()
        else:
            gaf_record["Date"] = properties["date"]

    def _add_gaf_evidence_code(self, gaf_record: dict, properties: Dict[str, str]) -> None:
        """Adds the GO evidence code to a GAF record"""
        if "evidence" not in properties:
            self.printer.print("Missing evidence code for feature '" + gaf_record["DB_Object_ID"] + "' and GO term '"
                               + gaf_record["GO_ID"] + "'")
            gaf_record["Evidence"] = self._default_evidence_code()
        else:
            evidence_code = self._back_convert_evidence_code(properties["evidence"])
            if not evidence_code:
                self.printer.print("Unrecognized evidence code: '" + properties["evidence"] + "' for feature '"
                                   + gaf_record["DB_Object_ID"] + "' and GO term '" + gaf_record["GO_ID"] + "'")
                gaf_record["Evidence"] = self._default_evidence_code()
            else:
                gaf_record["Evidence"] = evidence_code

    def _add_gaf_db_references(self, gaf_record: dict, publications: List[str], dbxrefs: List[str]) -> None:
        """Adds reference IDs to a GAF record"""
        gaf_record["DB:Reference"] = []
        for publication in publications:
            if publication != "null":
                gaf_record["DB:Reference"].append(publication)
        for dbxref in dbxrefs:
            db_authority = ontology.split_dbxref(dbxref)[0]
            if db_authority in self._db_reference_identifiers():
                gaf_record["DB:Reference"].append(dbxref)
        if not gaf_record["DB:Reference"]:
            gaf_record["DB:Reference"].append(self._default_db_reference())

    def _add_gaf_withfrom_info(self, gaf_record: dict, dbxrefs: List[str]) -> None:
        """Adds references for a GO evidence code to a GAF record"""
        gaf_record["With"] = []
        for dbxref in dbxrefs:
            db_authority = ontology.split_dbxref(dbxref)[0]
            if db_authority not in self._db_reference_identifiers():
                gaf_record["With"].append(dbxref)

    @staticmethod
    def _add_gaf_object_type(gaf_record: dict, featuretype: str) -> None:
        """Adds the feature type to a GAF record"""
        gaf_record["DB_Object_Type"] = featuretype

    @staticmethod
    def _add_gaf_object_name(gaf_record: dict, product_name: str) -> None:
        """Adds the name of a gene product to a GAF record"""
        gaf_record["DB_Object_Name"] = product_name

    @staticmethod
    def _add_gaf_object_symbol(gaf_record: dict, gene_name: str) -> None:
        """Adds the gene name to a GAF record"""
        if not gene_name:
            raise iobase.DatabaseError("Missing gene name for feature '" + gaf_record["DB_Object_ID"] + "'")
        gaf_record["DB_Object_Symbol"] = gene_name

    @staticmethod
    def _add_gaf_synonyms(gaf_record: dict, synonyms: List[str]) -> None:
        """Adds gene synonyms to a GAF record"""
        gaf_record["Synonym"] = synonyms



import unittest.mock
import sqlalchemy.schema
from .. import dbutils, utils, ddl


class RoleTests(unittest.TestCase):
    """Tests the handling of roles for a CHADO database"""

    connection_parameters = utils.parse_yaml(dbutils.default_configuration_file())
    connection_uri = dbutils.random_database_uri(connection_parameters)

    @classmethod
    def setUpClass(cls):
        # Creates a database and establishes a connection
        dbutils.create_database(cls.connection_uri)
        cls.client = ddl.RolesClient(cls.connection_uri)

    @classmethod
    def tearDownClass(cls):
        # Drops the database
        dbutils.drop_database(cls.connection_uri, True)

    def test_revoke_privileges_commands(self):
        # Tests the creation of commands for revoking privileges for database access
        commands = self.client.revoke_privileges_commands("testrole", "testschema")
        self.assertEqual(len(commands), 4)
        self.assertIn("REVOKE USAGE ON SCHEMA \"testschema\" FROM \"testrole\"", commands)
        self.assertIn("REVOKE ALL PRIVILEGES ON ALL TABLES IN SCHEMA \"testschema\" FROM \"testrole\"", commands)

    def test_grant_privileges_commands(self):
        # Tests the creation of commands for granting privileges for database access
        commands = self.client.grant_privileges_commands("testrole", "testschema", False)
        self.assertEqual(len(commands), 3)
        self.assertIn("GRANT USAGE ON SCHEMA \"testschema\" TO \"testrole\"", commands)
        self.assertNotIn("GRANT INSERT ON ALL TABLES IN SCHEMA \"testschema\" TO \"testrole\"", commands)

        commands = self.client.grant_privileges_commands("testrole", "testschema", True)
        self.assertEqual(len(commands), 8)
        self.assertIn("GRANT USAGE ON SCHEMA \"testschema\" TO \"testrole\"", commands)
        self.assertIn("GRANT INSERT ON ALL TABLES IN SCHEMA \"testschema\" TO \"testrole\"", commands)

        commands = self.client.grant_privileges_commands("testrole", "audit", True)
        self.assertEqual(len(commands), 6)

    @unittest.mock.patch("sqlalchemy.schema.DDL")
    def test_execute_ddl(self, mock_ddl):
        # Tests the function that executes DDL statements
        self.assertIs(mock_ddl, sqlalchemy.schema.DDL)
        self.client.execute_ddl(["command1", "command2"])
        self.assertIn(unittest.mock.call("command1"), mock_ddl.mock_calls)
        self.assertIn(unittest.mock.call("command2"), mock_ddl.mock_calls)
        self.assertIn(unittest.mock.call().execute(self.client.engine), mock_ddl.mock_calls)

        mock_ddl.reset_mock()
        self.client.execute_ddl("singlecommand")
        mock_ddl.assert_called_with("singlecommand")
        self.assertIn(unittest.mock.call().execute(self.client.engine), mock_ddl.mock_calls)

    @unittest.mock.patch("pychado.ddl.RolesClient.execute_ddl")
    @unittest.mock.patch("pychado.ddl.RolesClient.revoke_privileges_commands")
    @unittest.mock.patch("pychado.ddl.RolesClient.grant_privileges_commands")
    @unittest.mock.patch("pychado.ddl.RolesClient.schema_exists")
    @unittest.mock.patch("pychado.ddl.RolesClient.role_exists")
    def test_grant_or_revoke(self, mock_exists_role, mock_exists_schema, mock_grant, mock_revoke, mock_execute):
        # Tests the main function granting/revoking privileges
        self.assertIs(mock_exists_role, self.client.role_exists)
        self.assertIs(mock_exists_schema, self.client.schema_exists)
        self.assertIs(mock_grant, self.client.grant_privileges_commands)
        self.assertIs(mock_revoke, self.client.revoke_privileges_commands)
        self.assertIs(mock_execute, self.client.execute_ddl)

        mock_exists_role.return_value = False
        self.client.grant_or_revoke("testrole", "testschema", False, False)
        mock_exists_schema.assert_not_called()
        mock_execute.assert_not_called()

        mock_exists_role.return_value = True
        mock_exists_schema.return_value = False
        self.client.grant_or_revoke("testrole", "testschema", False, False)
        mock_exists_schema.assert_called()
        mock_execute.assert_not_called()

        mock_exists_role.return_value = True
        mock_exists_schema.return_value = True
        mock_grant.return_value = ["grant_command"]
        mock_revoke.return_value = ["revoke_command"]
        self.client.grant_or_revoke("testrole", "testschema", False, False)
        mock_revoke.assert_called_with("testrole", "testschema")
        mock_grant.assert_not_called()
        mock_execute.assert_called_with(["revoke_command"])

        self.client.grant_or_revoke("testrole", "testschema", False, True)
        mock_revoke.assert_called_with("testrole", "testschema")
        mock_grant.assert_called_with("testrole", "testschema", False)
        mock_execute.assert_called_with(["revoke_command", "grant_command"])


class SetupTests(unittest.TestCase):
    """Tests the setup of CHADO database schemas"""

    connection_parameters = utils.parse_yaml(dbutils.default_configuration_file())
    connection_uri = dbutils.random_database_uri(connection_parameters)

    @classmethod
    def setUpClass(cls):
        # Creates a database and establishes a connection
        dbutils.create_database(cls.connection_uri)
        cls.client = ddl.AuditSchemaSetupClient(cls.connection_uri)
        cls.public_metadata = sqlalchemy.schema.MetaData(schema='public')

    @classmethod
    def tearDownClass(cls):
        # Drops the database
        dbutils.drop_database(cls.connection_uri, True)

    def setUp(self):
        # Cleans the database
        sqlalchemy.schema.DDL("DROP SCHEMA public CASCADE").execute(self.client.engine)
        sqlalchemy.schema.DDL("CREATE SCHEMA public").execute(self.client.engine)
        if self.client.schema_exists("audit"):
            sqlalchemy.schema.DDL("DROP SCHEMA audit CASCADE").execute(self.client.engine)

    def test_create_trigger_function(self):
        # Tests the syntax of trigger function creation
        wrapper = self.client.create_trigger_function("trigger_schema", "trigger_name", "trigger_def")
        self.assertEqual(wrapper, "CREATE OR REPLACE FUNCTION trigger_schema.trigger_name()\n"
                                  "RETURNS trigger\n"
                                  "LANGUAGE plpgsql\n"
                                  "AS $function$\n"
                                  "BEGIN\n"
                                  "trigger_def;\n"
                                  "END;\n"
                                  "$function$")

    def test_create_generic_trigger(self):
        # Tests the syntax of trigger creation
        trigger = self.client.create_generic_trigger("testtrigger", "function_schema", "trigger_function",
                                                     "testschema", "testtable")
        self.assertEqual(trigger, "CREATE TRIGGER testtrigger AFTER INSERT OR UPDATE OR DELETE ON testschema.testtable "
                                  "FOR EACH ROW EXECUTE PROCEDURE function_schema.trigger_function()")

    def test_generic_audit_function(self):
        # Tests the syntax of an audit function
        fct = self.client.generic_audit_function("testschema", "testtable", ["col1", "col2"])
        self.assertEqual(fct, "IF TG_OP = 'INSERT' OR TG_OP = 'UPDATE' THEN\n"
                              "\tINSERT INTO testschema.testtable(type, col1, col2) "
                              "VALUES (CAST(TG_OP AS operation_type), NEW.col1, NEW.col2);\n"
                              "\tRETURN NEW;\n"
                              "ELSE\n"
                              "\tINSERT INTO testschema.testtable(type, col1, col2) "
                              "VALUES (CAST(TG_OP AS operation_type), OLD.col1, OLD.col2);\n"
                              "\tRETURN OLD;\n"
                              "END IF")

    def test_schema_exists(self):
        # Tests the function that checks if a schema exists
        res = self.client.schema_exists("testschema")
        self.assertFalse(res)
        sqlalchemy.schema.DDL("CREATE SCHEMA testschema").execute(self.client.engine)
        res = self.client.schema_exists("testschema")
        self.assertTrue(res)

    def test_table_exists(self):
        # Tests the function that checks if a table exists
        res = self.client.table_exists("testtable", "public")
        self.assertFalse(res)
        sqlalchemy.schema.DDL("CREATE TABLE testtable(id INTEGER, data VARCHAR(255))").execute(self.client.engine)
        res = self.client.table_exists("testtable", "public")
        self.assertTrue(res)

    def test_table_inherits(self):
        # Tests the function that checks if a table inherits from a parent table
        sqlalchemy.schema.DDL("CREATE TABLE capitals(state CHAR(2))").execute(self.client.engine)
        sqlalchemy.schema.DDL("CREATE TABLE cities(name TEXT, population INT)").execute(self.client.engine)
        res = self.client.table_inherits("capitals", "cities")
        self.assertFalse(res)
        sqlalchemy.schema.DDL("DROP TABLE capitals").execute(self.client.engine)
        sqlalchemy.schema.DDL("CREATE TABLE capitals(state CHAR(2)) INHERITS (cities)").execute(self.client.engine)
        res = self.client.table_inherits("capitals", "cities")
        self.assertTrue(res)

    def test_trigger_exists(self):
        # Tests the function that checks if a trigger exists
        sqlalchemy.schema.DDL("CREATE TABLE testtable(id INTEGER, data VARCHAR(255))").execute(self.client.engine)
        res = self.client.trigger_exists("public", "testtrigger")
        self.assertFalse(res)
        fct = self.client.create_trigger_function("public", "testfct", "SELECT 1+1")
        sqlalchemy.schema.DDL(fct).execute(self.client.engine)
        sqlalchemy.schema.DDL("CREATE TRIGGER testtrigger AFTER INSERT ON testtable EXECUTE PROCEDURE testfct()")\
            .execute(self.client.engine)
        res = self.client.trigger_exists("public", "testtrigger")
        self.assertTrue(res)

    def test_function_exists(self):
        # Tests the function that checks if a database function exists
        res = self.client.function_exists("pg_catalog", "now")
        self.assertTrue(res)
        res = self.client.function_exists("pg_catalog", "inexistent_function")
        self.assertFalse(res)
        res = self.client.function_exists("inexistent_schema", "now")
        self.assertFalse(res)

    def test_role_exists(self):
        # Tests the function that checks if a role exists
        res = self.client.role_exists("postgres")
        self.assertTrue(res)
        res = self.client.role_exists(utils.random_string(30))
        self.assertFalse(res)

    @unittest.mock.patch("pychado.ddl.AuditSchemaSetupClient.execute_ddl")
    @unittest.mock.patch("pychado.ddl.AuditSchemaSetupClient.schema_exists")
    def test_create_schema(self, mock_exists, mock_execute):
        # Tests the function that creates a schema
        self.assertIs(mock_exists, self.client.schema_exists)
        self.assertIs(mock_execute, self.client.execute_ddl)

        mock_exists.return_value = True
        self.client.create_schema()
        mock_exists.assert_called_with("audit")
        mock_execute.assert_not_called()

        mock_exists.return_value = False
        self.client.create_schema()
        mock_execute.assert_called_with("CREATE SCHEMA audit")

    @unittest.mock.patch("pychado.ddl.AuditSchemaSetupClient.execute_ddl")
    @unittest.mock.patch("pychado.ddl.AuditSchemaSetupClient.table_inherits")
    @unittest.mock.patch("pychado.ddl.AuditSchemaSetupClient.table_exists")
    def test_setup_inheritance(self, mock_exists, mock_inherits, mock_execute):
        # Tests the function that sets up table inheritance
        self.assertIs(mock_exists, self.client.table_exists)
        self.assertIs(mock_inherits, self.client.table_inherits)
        self.assertIs(mock_execute, self.client.execute_ddl)

        mock_exists.return_value = False
        mock_inherits.return_value = True
        self.client.setup_inheritance("cities", ["capitals"])
        mock_exists.assert_called_with("capitals", "audit")
        mock_inherits.assert_not_called()
        mock_execute.assert_not_called()

        mock_exists.return_value = True
        mock_inherits.return_value = True
        self.client.setup_inheritance("cities", ["capitals"])
        mock_exists.assert_called_with("capitals", "audit")
        mock_inherits.assert_called_with("audit.capitals", "audit.cities")
        mock_execute.assert_not_called()

        mock_exists.return_value = True
        mock_inherits.return_value = False
        self.client.setup_inheritance("cities", ["capitals"])
        mock_exists.assert_called_with("capitals", "audit")
        mock_inherits.assert_called_with("audit.capitals", "audit.cities")
        mock_execute.assert_called_with("ALTER TABLE audit.capitals INHERIT audit.cities")

    @unittest.mock.patch("pychado.ddl.AuditSchemaSetupClient.execute_ddl")
    @unittest.mock.patch("pychado.ddl.AuditSchemaSetupClient.trigger_exists")
    @unittest.mock.patch("pychado.ddl.AuditSchemaSetupClient.create_generic_trigger")
    @unittest.mock.patch("pychado.ddl.AuditSchemaSetupClient.create_trigger_function")
    @unittest.mock.patch("pychado.ddl.AuditSchemaSetupClient.generic_audit_function")
    def test_create_audit_triggers(self, mock_function, mock_wrapper, mock_trigger, mock_exists, mock_execute):
        # Tests the creation of audit trigger functions
        self.assertIs(mock_function, self.client.generic_audit_function)
        self.assertIs(mock_wrapper, self.client.create_trigger_function)
        self.assertIs(mock_trigger, self.client.create_generic_trigger)
        self.assertIs(mock_exists, self.client.trigger_exists)
        self.assertIs(mock_execute, self.client.execute_ddl)

        testtable = sqlalchemy.Table("testtable", self.client.metadata,
                                     sqlalchemy.Column("idcolumn", sqlalchemy.INTEGER, primary_key=True),
                                     sqlalchemy.Column("datacolumn", sqlalchemy.TEXT))

        mock_function.return_value = "function_definition"
        mock_wrapper.return_value = "function_wrapper"
        mock_trigger.return_value = "trigger_wrapper"
        mock_exists.return_value = True
        self.client.create_audit_triggers([testtable])
        mock_function.assert_called_with("audit", "testtable", ["idcolumn", "datacolumn"])
        mock_wrapper.assert_called_with("audit", "public_testtable_proc", "function_definition")
        mock_trigger.assert_called_with("testtable_audit_tr", "audit", "public_testtable_proc", "public", "testtable")
        mock_execute.assert_called_with("function_wrapper")

        mock_execute.reset_mock()
        mock_exists.return_value = False
        self.client.create_audit_triggers([testtable])
        mock_execute.assert_has_calls([unittest.mock.call("function_wrapper"),
                                       unittest.mock.call("trigger_wrapper")])

    def test_create_audit_tables(self):
        # Tests the creation of the audit tables
        newtable = sqlalchemy.Table(
            "newtable", self.public_metadata,
            sqlalchemy.Column("idcolumn", sqlalchemy.INTEGER, nullable=False),
            sqlalchemy.Column("datacolumn", sqlalchemy.TEXT, nullable=True, server_default="whatever"))
        audit_tables = self.client.create_audit_tables([newtable])
        self.assertEqual(len(audit_tables), 1)
        audit_table = audit_tables[0]
        self.assertEqual(len(audit_table.columns), 6)
        for column in audit_table.columns:
            if column.name == "idcolumn":
                self.assertFalse(column.primary_key)
            elif column.name == "audit_id":
                self.assertTrue(column.primary_key)


class BackupSetupTests(unittest.TestCase):

    connection_parameters = utils.parse_yaml(dbutils.default_configuration_file())
    connection_uri = dbutils.random_database_uri(connection_parameters)

    @classmethod
    def setUpClass(cls):
        # Creates a database and establishes a connection
        dbutils.create_database(cls.connection_uri)
        cls.client = ddl.AuditBackupSchemaSetupClient(cls.connection_uri)
        cls.public_metadata = sqlalchemy.schema.MetaData(schema='public')

    @classmethod
    def tearDownClass(cls):
        # Drops the database
        dbutils.drop_database(cls.connection_uri, True)

    @unittest.mock.patch("sqlalchemy.engine.Connection.execute")
    @unittest.mock.patch("pychado.ddl.AuditBackupSchemaSetupClient.function_exists")
    def test_execute_backup_function(self, mock_exists: unittest.mock.Mock, mock_execute: unittest.mock.Mock):
        # Tests the function that executes the backup of the audit schema
        self.assertIs(mock_exists, self.client.function_exists)
        self.assertIs(mock_execute, self.client.connection.execute)

        mock_exists.return_value = False
        self.client.execute_backup_function("testdate")
        mock_execute.assert_not_called()

        mock_exists.return_value = True
        self.client.execute_backup_function("testdate")
        self.assertIn(unittest.mock.call().scalar(), mock_execute.mock_calls)

    def test_backup_function_wrapper(self):
        # Tests the syntax of the backup function
        fct = self.client.backup_function_wrapper("testname", ["var integer"], "testdefinition")
        self.assertEqual(fct, "CREATE OR REPLACE FUNCTION testname(text)\n"
                              "RETURNS bigint\n"
                              "LANGUAGE plpgsql\n"
                              "AS $function$\n"
                              "DECLARE\n"
                              "var integer;\n"
                              "BEGIN\n"
                              "testdefinition;\n"
                              "END;\n"
                              "$function$")


if __name__ == '__main__':
    unittest.main(verbosity=2, buffer=True)



import sqlalchemy.orm
from . import base, general, cv


# Object-relational mappings for the CHADO Organism module


class Organism(base.PublicBase):
    """Class for the CHADO 'organism' table"""
    # Columns
    organism_id = sqlalchemy.Column(sqlalchemy.BIGINT, nullable=False, primary_key=True, autoincrement=True)
    abbreviation = sqlalchemy.Column(sqlalchemy.VARCHAR(255), nullable=False)
    genus = sqlalchemy.Column(sqlalchemy.VARCHAR(255), nullable=False)
    species = sqlalchemy.Column(sqlalchemy.VARCHAR(255), nullable=False)
    common_name = sqlalchemy.Column(sqlalchemy.VARCHAR(255), nullable=True)
    infraspecific_name = sqlalchemy.Column(sqlalchemy.VARCHAR(1024), nullable=True)
    type_id = sqlalchemy.Column(sqlalchemy.BIGINT, sqlalchemy.ForeignKey(
        cv.CvTerm.cvterm_id, onupdate="CASCADE", ondelete="CASCADE"), nullable=True)
    comment = sqlalchemy.Column(sqlalchemy.TEXT, nullable=True)
    version = sqlalchemy.orm.deferred(sqlalchemy.Column(sqlalchemy.INTEGER, nullable=True))

    # Constraints
    __tablename__ = "organism"
    __table_args__ = (sqlalchemy.UniqueConstraint(genus, species, infraspecific_name, "version", name="organism_c1"),
                      sqlalchemy.UniqueConstraint(abbreviation, name="organism_c2"))

    # Relationships
    type = sqlalchemy.orm.relationship(cv.CvTerm, foreign_keys=type_id, backref="organism_type")

    # Initialisation
    def __init__(self, genus, species, abbreviation=None, common_name=None, infraspecific_name=None,
                 type_id=None, comment=None, version=None, organism_id=None):
        for key, value in locals().items():
            if key != self:
                setattr(self, key, value)

    # Representation
    def __repr__(self):
        return "<organism.Organism(organism_id={0}, abbreviation='{1}', genus='{2}', species='{3}', " \
               "common_name='{4}', infraspecific_name='{5}', type_id={6}, comment='{7}')>"\
            .format(self.organism_id, self.abbreviation, self.genus, self.species, self.common_name,
                    self.infraspecific_name, self.type_id, self.comment)


class OrganismDbxRef(base.PublicBase):
    """Class for the CHADO 'organism_dbxref' table"""
    # Columns
    organism_dbxref_id = sqlalchemy.Column(sqlalchemy.BIGINT, nullable=False, primary_key=True, autoincrement=True)
    organism_id = sqlalchemy.Column(sqlalchemy.BIGINT, sqlalchemy.ForeignKey(
        Organism.organism_id, onupdate="CASCADE", ondelete="CASCADE"), nullable=False)
    dbxref_id = sqlalchemy.Column(sqlalchemy.BIGINT, sqlalchemy.ForeignKey(
        general.DbxRef.dbxref_id, onupdate="CASCADE", ondelete="CASCADE"), nullable=False)

    # Constraints
    __tablename__ = "organism_dbxref"
    __table_args__ = (sqlalchemy.UniqueConstraint(organism_id, dbxref_id, name="organism_dbxref_c1"),
                      sqlalchemy.Index("organism_dbxref_idx1", organism_id),
                      sqlalchemy.Index("organism_dbxref_idx2", dbxref_id))

    # Relationships
    organism = sqlalchemy.orm.relationship(Organism, foreign_keys=organism_id, backref="organism_dbxref_organism")
    dbxref = sqlalchemy.orm.relationship(general.DbxRef, backref="organism_dbxref_dbxref")

    # Initialisation
    def __init__(self, organism_id, dbxref_id, organism_dbxref_id=None):
        for key, value in locals().items():
            if key != self:
                setattr(self, key, value)

    # Representation
    def __repr__(self):
        return "<organism.OrganismDbxRef(organism_dbxref_id={0}, organism_id={1}, dbxref_id={2})>".format(
            self.organism_dbxref_id, self.organism_id, self.dbxref_id)


class OrganismProp(base.PublicBase):
    """Class for the CHADO 'organismprop' table"""
    # Columns
    organismprop_id = sqlalchemy.Column(sqlalchemy.BIGINT, nullable=False, primary_key=True, autoincrement=True)
    organism_id = sqlalchemy.Column(sqlalchemy.BIGINT, sqlalchemy.ForeignKey(
        Organism.organism_id, onupdate="CASCADE", ondelete="CASCADE"), nullable=False)
    type_id = sqlalchemy.Column(sqlalchemy.BIGINT, sqlalchemy.ForeignKey(
        cv.CvTerm.cvterm_id, onupdate="CASCADE", ondelete="CASCADE"), nullable=False)
    value = sqlalchemy.Column(sqlalchemy.TEXT, nullable=True)
    rank = sqlalchemy.Column(sqlalchemy.INTEGER, nullable=False, server_default="0")

    # Constraints
    __tablename__ = "organismprop"
    __table_args__ = (sqlalchemy.UniqueConstraint(organism_id, type_id, rank, name="organismprop_c1"),
                      sqlalchemy.Index("organismprop_idx1", organism_id),
                      sqlalchemy.Index("organismprop_idx2", type_id))

    # Relationships
    organism = sqlalchemy.orm.relationship(Organism, foreign_keys=organism_id, backref="organismprop_organism")
    type = sqlalchemy.orm.relationship(cv.CvTerm, foreign_keys=type_id, backref="organismprop_type")

    # Initialisation
    def __init__(self, organism_id, type_id, value=None, rank=0, organismprop_id=None):
        for key, val in locals().items():
            if key != self:
                setattr(self, key, val)

    # Representation
    def __repr__(self):
        return "<organism.OrganismProp(organismprop_id={0}, organism_id={1}, type_id={2}, value='{3}', rank={4})>"\
            .format(self.organismprop_id, self.organism_id, self.type_id, self.value, self.rank)



import unittest.mock
from .. import dbutils, utils
from ..io import direct
from ..orm import base, organism, cv


class TestDirectIO(unittest.TestCase):
    """Tests various functions used to insert, update or delete entries in database tables"""

    connection_parameters = utils.parse_yaml(dbutils.default_configuration_file())
    connection_uri = dbutils.random_database_uri(connection_parameters)

    @classmethod
    def setUpClass(cls):
        # Creates a database, establishes a connection and creates tables
        dbutils.create_database(cls.connection_uri)
        cls.client = direct.DirectIOClient(cls.connection_uri)
        schema_base = base.PublicBase
        schema_metadata = schema_base.metadata
        schema_base.metadata.create_all(cls.client.engine, tables=schema_metadata.sorted_tables)

    @classmethod
    def tearDownClass(cls):
        # Drops the database
        dbutils.drop_database(cls.connection_uri, True)

    def tearDown(self):
        # Rolls back all changes to the database
        self.client.session.rollback()

    def test_insert_delete_organism(self):
        # Tests the insertion of an organism into a Chado database and its deletion

        # Check the table is initially empty
        zero_result = self.client.query_table(organism.Organism).first()
        self.assertIsNone(zero_result)

        # Insert an organism and check that the insertion is successful
        self.client.insert_organism('AAA', 'BBB', 'CCC', 'DDD', 'EEE', 'FFF')
        first_result = self.client.query_table(organism.Organism).all()
        self.assertEqual(len(first_result), 1)
        self.assertEqual(first_result[0].genus, 'AAA')

        # Try to insert an organism with the same properties and check that nothing happened
        self.client.insert_organism('AAA', 'BBB', 'CCC', 'DDD', 'EEE', 'FFF')
        second_result = self.client.query_table(organism.Organism).all()
        self.assertEqual(first_result, second_result)

        # Try to delete an organism that doesn't exist and check that nothing happened
        self.client.delete_organism('XXX')
        third_result = self.client.query_table(organism.Organism).all()
        self.assertEqual(first_result, third_result)

        # Delete the previously inserted organism and check that it is gone
        self.client.delete_organism('CCC')
        fifth_result = self.client.query_table(organism.Organism).first()
        self.assertIsNone(fifth_result)

    @unittest.mock.patch("pychado.io.direct.DirectIOClient.query_organisms_by_property_type")
    @unittest.mock.patch("pychado.io.direct.DirectIOClient.query_all_organisms")
    @unittest.mock.patch("pychado.io.direct.DirectIOClient._load_cvterm")
    def test_select_organisms(self, mock_load_cvterm: unittest.mock.Mock, mock_query_all: unittest.mock.Mock,
                              mock_query_public: unittest.mock.Mock):
        # Tests the selection of organisms from a Chado database
        self.assertIs(mock_load_cvterm, self.client._load_cvterm)
        self.assertIs(mock_query_all, self.client.query_all_organisms)
        self.assertIs(mock_query_public, self.client.query_organisms_by_property_type)
        mock_load_cvterm.return_value = cv.CvTerm(cv_id=1, dbxref_id=2, cvterm_id=3, name="")

        self.client.select_organisms(False, False)
        mock_load_cvterm.assert_not_called()
        mock_query_public.assert_not_called()
        mock_query_all.assert_called()

        mock_load_cvterm.reset_mock()
        mock_query_public.reset_mock()
        mock_query_all.reset_mock()
        self.client.select_organisms(True, False)
        mock_load_cvterm.assert_called_with("genedb_public")
        mock_query_public.assert_called_with(3, False)
        mock_query_all.assert_not_called()



import sqlalchemy.sql
from . import iobase
from ..orm import organism


class DirectIOClient(iobase.ChadoClient):
    """Class for inserting, updating or deleting database entries"""

    def select_organisms(self, public_only: bool, query_version: bool) -> sqlalchemy.sql.Select:
        """Loads organisms from the database"""
        if public_only:
            public_type_term = self._load_cvterm("genedb_public")
            query = self.query_organisms_by_property_type(public_type_term.cvterm_id, query_version)
        else:
            query = self.query_all_organisms(query_version)
        return query.statement

    def insert_organism(self, genus: str, species: str, abbreviation: str, common_name=None,
                        infraspecific_name=None, comment=None, genome_version=None):
        """Inserts an organism into the database"""
        genus_species_entry = self.session.query(organism.Organism).filter_by(
            genus=genus, species=species, infraspecific_name=infraspecific_name).first()
        abbreviation_entry = self.session.query(organism.Organism).filter_by(abbreviation=abbreviation).first()
        if genus_species_entry:
            print("An organism with genus '" + genus + "', species '" + species
                  + "' and strain '" + infraspecific_name + "' is already present in the database.")
        elif abbreviation_entry:
            print("An organism with abbreviation '" + abbreviation + "' is already present in the database.")
        else:
            organism_entry = organism.Organism(genus=genus, species=species, abbreviation=abbreviation,
                                               common_name=common_name, infraspecific_name=infraspecific_name,
                                               comment=comment, version=genome_version)
            self.session.add(organism_entry)
            self.session.commit()
            print("An organism with genus '" + genus + "' and species '" + species
                  + "' has been inserted into the database.")

    def delete_organism(self, abbreviation: str):
        """Deletes an organism from the database"""
        obj = self.session.query(organism.Organism).filter_by(abbreviation=abbreviation).first()
        if not obj:
            print("An organism with abbreviation '" + abbreviation + "' is not present in the database.")
        else:
            self.session.delete(obj)
            self.session.commit()
            print("An organism with abbreviation '" + abbreviation + "' has been deleted from the database.")



from . import utils, dbutils, queries, ddl
from .io import direct, essentials, ontology, fasta, gff, gaf


def create_connection_string(filename: str, dbname: str) -> str:
    """Reads database connection parameters from a configuration file and generates a connection string"""
    if not filename:
        filename = dbutils.default_configuration_file()
    connection_parameters = utils.parse_yaml(filename)
    connection_parameters["database"] = dbname
    return dbutils.generate_uri(connection_parameters)


def check_access(connection_uri: str, task: str) -> bool:
    """Checks if the database of interest exists and is accessible. If the database doesn't exist, but the
    task implies its creation, create it. Otherwise exit the program."""
    exists = dbutils.exists(connection_uri)
    if exists:
        if task in ["create", "restore"]:
            # Database already exists, we should not overwrite it. Return without further action
            print("Database already exists. Overwriting is not permitted.")
            return False
        else:
            # Database exists, that's all we need
            return True
    else:
        if task in ["create", "restore"]:
            # Database doesn't exist, but task implies its creation
            return True
        else:
            # Database doesn't exist, and task can't be completed. Return without further action
            print("Database does not exist. Task can't be completed.")
            return False


def init(command: str) -> None:
    """Initiates or resets the default connection parameters"""
    if command == "init":
        # Set the default connection parameters
        dbutils.set_default_parameters()
    elif command == "reset":
        # Reset the default connection parameters to factory settings
        dbutils.reset_default_parameters()
    else:
        print("Functionality '" + command + "' is not yet implemented.")


def run_command_with_arguments(command: str, sub_command: str, arguments, connection_uri: str) -> None:
    """Runs a specified sub-command with the supplied arguments"""

    # Run the command
    if command == "connect":
        # Connect to a PostgreSQL database for an interactive session
        dbutils.connect_to_database(connection_uri)
    elif command == "admin" and sub_command == "create":
        # Create a PostgreSQL database
        dbutils.create_database(connection_uri)
    elif command == "admin" and sub_command == "drop":
        # Drop a PostgreSQL database
        dbutils.drop_database(connection_uri)
    elif command == "admin" and sub_command == "dump":
        # Dump a PostgreSQL database into an archive file
        dbutils.dump_database(connection_uri, arguments.archive)
    elif command == "admin" and sub_command == "restore":
        # Restore a PostgreSQL database from an archive file
        dbutils.create_database(connection_uri)
        dbutils.restore_database(connection_uri, arguments.archive)
    elif command == "admin" and sub_command == "setup":
        # Setup a PostgreSQL database according to a schema
        run_setup_command(arguments, connection_uri)
    elif command == "admin" and sub_command == "grant":
        # Grant access to objects in a PostgreSQL database
        run_grant_revoke_command(arguments, connection_uri, True)
    elif command == "admin" and sub_command == "revoke":
        # Revoke access to objects in a PostgreSQL database
        run_grant_revoke_command(arguments, connection_uri, False)
    elif command == "query":
        # Query a PostgreSQL database and export the result to a text file
        run_query_command(arguments, connection_uri)
    elif command == "execute":
        # Run a function defined in a PostgreSQL database
        run_execute_command(sub_command, arguments, connection_uri)
    elif command == "extract":
        # Run a pre-compiled query against the CHADO database
        run_select_command(sub_command, arguments, connection_uri)
    elif command == "insert":
        # Insert a new entity of a specified type into the CHADO database
        run_insert_command(sub_command, arguments, connection_uri)
    elif command == "delete":
        # Delete an entity of a specified type from the CHADO database
        run_delete_command(sub_command, arguments, connection_uri)
    elif command == "import":
        # Import data from file into the CHADO database
        run_import_command(sub_command, arguments, connection_uri)
    elif command == "export":
        # Export data from the CHADO database to file
        run_export_command(sub_command, arguments, connection_uri)
    else:
        print("Functionality '" + command + "' is not yet implemented.")


def run_setup_command(arguments, uri: str) -> None:
    """Sets up a PostgreSQL database according to a schema"""
    if arguments.schema_file or arguments.schema == "gmod":
        schema_file = arguments.schema_file
        if not schema_file:
            schema_file = utils.download_file(dbutils.default_schema_url())
        dbutils.setup_database(uri, schema_file)
    else:
        if arguments.schema == "basic":
            client = ddl.PublicSchemaSetupClient(uri)
        elif arguments.schema == "audit":
            client = ddl.AuditSchemaSetupClient(uri)
        elif arguments.schema == "audit_backup":
            client = ddl.AuditBackupSchemaSetupClient(uri)
        else:
            client = ddl.DDLClient(uri)
        client.create()


def run_grant_revoke_command(arguments, uri: str, grant_access: bool) -> None:
    """Grant/revoke access to objects in a PostgreSQL database"""
    client = ddl.RolesClient(uri)
    if grant_access:
        client.grant_or_revoke(arguments.role, arguments.schema, arguments.write, True)
    else:
        client.grant_or_revoke(arguments.role, arguments.schema, False, False)


def run_query_command(arguments, uri: str) -> None:
    """Query a PostgreSQL database and export the result to a text file"""
    query = (arguments.query or utils.read_text(arguments.input_file))
    dbutils.query_and_print(uri, query, arguments.output_file, arguments.format, arguments.include_header,
                            arguments.delimiter)
    if arguments.output_file:
        print("Data exported to " + arguments.output_file)


def run_execute_command(specifier: str, arguments, uri: str) -> None:
    """Run a function defined in the database"""
    if specifier == "audit_backup":
        client = ddl.AuditBackupSchemaSetupClient(uri)
        client.execute_backup_function(arguments.date)
    else:
        print("Functionality 'execute " + specifier + "' is not yet implemented.")


def run_select_command(specifier: str, arguments, uri: str) -> None:
    """Run a pre-compiled query against a database"""
    if specifier == "organisms":
        client = direct.DirectIOClient(uri)
        query = client.select_organisms(arguments.public_only, arguments.extract_version)
    else:
        template = queries.load_query(specifier)
        if specifier == "cvterms":
            query = queries.set_query_conditions(template, database=arguments.database, vocabulary=arguments.vocabulary)
        elif specifier == "genedb_products":
            query = queries.set_query_conditions(template, organism=arguments.organism)
        elif specifier == "stats":
            query = queries.set_query_conditions(template, organism=arguments.organism, start_date=arguments.start_date,
                                                 end_date=(arguments.end_date or utils.current_date()))
        else:
            print("Functionality 'extract " + specifier + "' is not yet implemented.")
            query = queries.set_query_conditions("")
    dbutils.query_and_print(uri, query, arguments.output_file, arguments.format, arguments.include_header,
                            arguments.delimiter)
    if arguments.output_file:
        print("Data exported to " + arguments.output_file)


def run_insert_command(specifier: str, arguments, uri: str) -> None:
    """Insert a new entity of a specified type into a database"""
    client = direct.DirectIOClient(uri)
    if specifier == "organism":
        client.insert_organism(arguments.genus, arguments.species, arguments.abbreviation, arguments.common_name,
                               arguments.infraspecific_name, arguments.comment, arguments.genome_version)
    else:
        print("Functionality 'insert " + specifier + "' is not yet implemented.")


def run_delete_command(specifier: str, arguments, uri: str) -> None:
    """Delete an entity of a specified type from a database"""
    client = direct.DirectIOClient(uri)
    if specifier == "organism":
        client.delete_organism(arguments.organism)
    else:
        print("Functionality 'delete " + specifier + "' is not yet implemented.")


def run_import_command(specifier: str, arguments, uri: str) -> None:
    """Imports data from a file into a database"""
    file = None
    if hasattr(arguments, "input_file") and arguments.input_file:
        file = arguments.input_file
    elif hasattr(arguments, "input_url") and arguments.input_url:
        file = utils.download_file(arguments.input_url)

    if specifier == "essentials":
        client = essentials.EssentialsClient(uri, arguments.verbose)
        client.load()
    elif specifier == "ontology":
        client = ontology.OntologyClient(uri, arguments.verbose)
        client.load(file, arguments.format, arguments.database_authority)
    elif specifier == "gff":
        client = gff.GFFImportClient(uri, arguments.verbose)
        client.load(file, arguments.organism, arguments.fasta, arguments.sequence_type, arguments.fresh_load,
                    arguments.force, arguments.full_genome)
    elif specifier == "fasta":
        client = fasta.FastaImportClient(uri, arguments.verbose)
        client.load(file, arguments.organism, arguments.sequence_type)
    elif specifier == "gaf":
        client = gaf.GAFImportClient(uri, arguments.verbose)
        client.load(file, arguments.organism, arguments.annotation_level)
    else:
        print("Functionality 'import " + specifier + "' is not yet implemented.")


def run_export_command(specifier: str, arguments, uri: str) -> None:
    """Exports data from a database to a file"""
    if specifier == "fasta":
        client = fasta.FastaExportClient(uri, arguments.verbose)
        client.export(arguments.output_file, arguments.organism, arguments.sequence_type, arguments.release,
                      arguments.extract_version, arguments.include_obsolete)
    elif specifier == "gff":
        client = gff.GFFExportClient(uri, arguments.verbose)
        client.export(arguments.output_file, arguments.organism, arguments.export_fasta, arguments.fasta_file,
                      arguments.include_obsolete)
    elif specifier == "gaf":
        client = gaf.GAFExportClient(uri, arguments.verbose)
        client.export(arguments.output_file, arguments.organism, arguments.database_authority,
                      arguments.annotation_level, arguments.include_obsolete)
    else:
        print("Functionality 'export " + specifier + "' is not yet implemented.")



import os
import unittest
import pronto
from .. import dbutils, utils
from ..io import iobase, essentials, ontology
from ..orm import base, general, cv


class TestOntology(unittest.TestCase):
    """Tests various functions used to load an ontology from a file into a database"""

    modules_dir = os.path.dirname(os.path.abspath(ontology.__file__))
    data_dir = os.path.join(modules_dir, '..', 'tests', 'data')
    connection_parameters = utils.parse_yaml(dbutils.default_configuration_file())
    connection_uri = dbutils.random_database_uri(connection_parameters)

    @classmethod
    def setUpClass(cls):
        # Creates a database, establishes a connection, creates tables and populates them with essential entries
        dbutils.create_database(cls.connection_uri)
        schema_base = base.PublicBase
        schema_metadata = schema_base.metadata
        essentials_client = essentials.EssentialsClient(cls.connection_uri)
        schema_metadata.create_all(essentials_client.engine, tables=schema_metadata.sorted_tables)
        essentials_client.load()
        essentials_client._load_further_relationship_entries()
        cls.client = ontology.OntologyClient(cls.connection_uri)

    @classmethod
    def tearDownClass(cls):
        # Drops the database
        dbutils.drop_database(cls.connection_uri, True)

    def setUp(self):
        # Inserts default entries into database tables
        (self.default_db, self.default_dbxref, self.default_cv, self.default_cvterm) = self.insert_default_entries()

    def tearDown(self):
        # Rolls back all changes to the database
        self.client.session.rollback()

    def insert_default_entries(self):
        # Inserts CV terms needed as basis for virtually all tests
        default_db = general.Db(name="defaultdb")
        self.client.add_and_flush(default_db)
        default_dbxref = general.DbxRef(db_id=default_db.db_id, accession="defaultaccession")
        self.client.add_and_flush(default_dbxref)
        default_cv = cv.Cv(name="defaultcv")
        self.client.add_and_flush(default_cv)
        default_cvterm = cv.CvTerm(cv_id=default_cv.cv_id, dbxref_id=default_dbxref.dbxref_id, name="testterm")
        self.client.add_and_flush(default_cvterm)
        return default_db, default_dbxref, default_cv, default_cvterm

    def test_split_dbxref(self):
        # Checks the splitting of a given database cross reference into its constituents
        result = ontology.split_dbxref("testdb:testaccession:testversion")
        self.assertEqual(result[0], "testdb")
        self.assertEqual(result[1], "testaccession")
        self.assertEqual(result[2], "testversion")

        result = ontology.split_dbxref("testdb:testaccession")
        self.assertEqual(result[0], "testdb")
        self.assertEqual(result[1], "testaccession")
        self.assertEqual(result[2], "")

        with self.assertRaises(AttributeError):
            ontology.split_dbxref("testdb_testaccession")

    def test_create_dbxref(self):
        # Checks the creation of a database cross reference from its constituents
        result = ontology.create_dbxref("testdb", "testaccession", "testversion")
        self.assertEqual(result, "testdb:testaccession:testversion")

        result = ontology.create_dbxref("testdb", "testaccession")
        self.assertEqual(result, "testdb:testaccession")

        with self.assertRaises(AttributeError):
            ontology.create_dbxref("testdb", "")

    def test_create_cvterm_entry(self):
        # Checks if an ontology term is correctly converted into a CV term entry
        term = pronto.Term(id="testid", name="testname", desc="testdescription", other={"is_obsolete": ["True"]})
        cvterm_entry = ontology.create_cvterm_entry(term, 100, 200)
        self.assertEqual(cvterm_entry, cv.CvTerm(cv_id=100, dbxref_id=200, name="testname",
                                                 definition="testdescription", is_obsolete=1))
        term = pronto.Term(id="otherid", name="othername", other={"is_obsolete": ["False"]})
        cvterm_entry = ontology.create_cvterm_entry(term, 10, 20)
        self.assertEqual(cvterm_entry, cv.CvTerm(cv_id=10, dbxref_id=20, name="othername", is_obsolete=0))

    def test_update_cvterm_properties(self):
        # Checks if the properties of a CV term are correctly updated
        input_term = cv.CvTerm(cv_id=100, dbxref_id=200, name="testname", definition="testdescription", is_obsolete=1)
        output_term = cv.CvTerm(cv_id=1, dbxref_id=200, name="newname")
        self.assertNotEqual(input_term, output_term)
        self.assertTrue(ontology.update_cvterm_properties(output_term, input_term))
        self.assertEqual(input_term, output_term)
        self.assertFalse(ontology.update_cvterm_properties(output_term, input_term))

    def test_ontology_parser(self):
        # checks if an ontology file is parsed correctly
        filename = os.path.join(self.data_dir, "io_obo_example.obo")
        content = ontology.parse_ontology(filename)
        self.assertEqual(len(content), 2)
        self.assertIn("test:0000001", content)
        term = content["test:0000001"]
        self.assertEqual(term.name, "diplodocus")
        self.assertEqual(term.desc, "definition of a diplodocus")
        self.assertEqual(term.other["namespace"][0], "animals")
        self.assertEqual(len(term.relations), 1)

    def test_get_default_namespace(self):
        # Checks if the default namespace is correctly extracted from an ontology
        ont = pronto.Ontology()
        namespace = ontology.get_default_namespace(ont)
        self.assertEqual(namespace, "")
        ont.meta["default-namespace"] = ["mynamespace"]
        namespace = ontology.get_default_namespace(ont)
        self.assertEqual(namespace, "mynamespace")

    def test_filter_ontology(self):
        # Checks the filtering of ontology terms by a database authority
        ont = pronto.Ontology()
        ont.include(pronto.Term("test:001:abc"))
        ont.include(pronto.Term("test:002"))
        ont.include(pronto.Term("other:001"))
        filtered_ontology = ontology.filter_ontology_by_db(ont, "test")
        self.assertEqual(len(filtered_ontology), 2)
        self.assertIn("test:001:abc", filtered_ontology)
        self.assertEqual(filtered_ontology["test:002"].id, "test:002")

    def test_extract_comments(self):
        # Tests the extraction of comments from an ontology term
        term = pronto.Term("testid")
        comment = ontology.extract_comment(term)
        self.assertEqual(comment, "")
        term.other = {"comment": ["testcomment"]}
        comment = ontology.extract_comment(term)
        self.assertEqual(comment, "testcomment")

    def test_extract_synonyms(self):
        # Tests the extraction of synonyms from an ontology term
        term = pronto.Term("testid")
        synonyms = ontology.extract_synonyms(term)
        self.assertEqual(len(synonyms), 0)
        term.synonyms = {pronto.Synonym("first_synonym"), pronto.Synonym("second_synonym")}
        synonyms = ontology.extract_synonyms(term)
        self.assertEqual(len(synonyms), 2)
        self.assertIn("second_synonym", synonyms)

    def test_extract_cross_references(self):
        # Tests the extraction of cross references from an ontology term
        term = pronto.Term("testid")
        crossrefs = ontology.extract_cross_references(term)
        self.assertEqual(len(crossrefs), 0)
        term.other = {"alt_id": ["alternative_id"], "xref": ["first_ref", "second_ref"]}
        crossrefs = ontology.extract_cross_references(term)
        self.assertEqual(len(crossrefs), 3)
        self.assertIn("alternative_id", crossrefs)
        self.assertIn("second_ref", crossrefs)

    def test_load_comment_term(self):
        # Tests the loading of the "comment" term from the database
        comment_term = self.client._load_comment_term()
        self.assertEqual(comment_term.name, "comment")

    def test_load_relationship_terms(self):
        # Tests the loading of relationship terms from the database
        relationship_terms = self.client._load_relationship_terms()
        self.assertIn("is_a", relationship_terms)
        self.assertEqual(relationship_terms["is_a"].is_relationshiptype, 1)

    def test_load_synonym_type_terms(self):
        # Tests the loading of synonym type terms from the database
        synonym_type_terms = self.client._load_synonym_type_terms()
        self.assertIn("exact", synonym_type_terms)

    def test_handle_db(self):
        # Tests the insertion of DB entries

        # Insert a new entry
        first_db = self.client._handle_db("testdb")
        self.assertEqual(first_db.name, "testdb")
        self.assertIsNotNone(first_db.db_id)

        # Insert a further entry with a different authority
        second_db = self.client._handle_db("anotherdb")
        self.assertNotEqual(first_db.db_id, second_db.db_id)

        # Try to insert a further entry with the same authority, and check that the function returns the existing entry
        third_db = self.client._handle_db("testdb")
        self.assertEqual(third_db.db_id, first_db.db_id)

    def test_handle_cv(self):
        # Tests the insertion of CV entries

        # Insert a new entry
        term = pronto.Term("testdb:testaccession")
        first_cv = self.client._handle_cv(term, "testnamespace")
        self.assertEqual(first_cv.name, "testnamespace")
        self.assertIsNotNone(first_cv.cv_id)

        # Try to insert an entry without namespace, and fail
        with self.assertRaises(iobase.InputFileError):
            self.client._handle_cv(term, "")

        # Insert a further entry with another namespace
        term.other = {"namespace": ["othernamespace"]}
        second_cv = self.client._handle_cv(term, "")
        self.assertEqual(second_cv.name, "othernamespace")
        self.assertNotEqual(first_cv.cv_id, second_cv.cv_id)

        # Try to insert an entry with the same namespace, and check that the function returns the existing entry
        third_cv = self.client._handle_cv(term, "")
        self.assertEqual(third_cv.cv_id, second_cv.cv_id)

    def test_handle_dbxref(self):
        # Tests the insertion and update of dbxref entries

        # Insert a new entry
        term = pronto.Term("defaultdb:testaccession")
        first_dbxref = self.client._handle_dbxref(term, self.default_db)
        self.assertEqual(first_dbxref.accession, "testaccession")
        self.assertEqual(first_dbxref.version, "")
        self.assertIsNotNone(first_dbxref.dbxref_id)

        # Insert a further entry with another accession
        term.id = "defaultdb:otheraccession"
        second_dbxref = self.client._handle_dbxref(term, self.default_db)
        self.assertEqual(second_dbxref.accession, "otheraccession")
        self.assertNotEqual(first_dbxref.dbxref_id, second_dbxref.dbxref_id)

        # Try to insert an entry with the same accession, but other version, and check that the function returns
        # the existing entry with updated version
        term.id = "defaultdb:testaccession:testversion"
        third_dbxref = self.client._handle_dbxref(term, self.default_db)
        self.assertEqual(third_dbxref.accession, "testaccession")
        self.assertEqual(third_dbxref.version, "testversion")
        self.assertEqual(first_dbxref.dbxref_id, third_dbxref.dbxref_id)

    def test_is_cvterm_unique(self):
        # Tests the functionality that checks if a CV term is unique in a database

        # Test a CV term with same dbxref and same properties - should be unique
        (is_unique, term, dbxref) = self.client._is_cvterm_unique(cv.CvTerm(
            cv_id=self.default_cv.cv_id, dbxref_id=self.default_dbxref.dbxref_id, name="testterm", is_obsolete=0))
        self.assertTrue(is_unique)
        self.assertIsNotNone(term)

        # Test a CV term with same properties, but different dbxref - should not be unique
        (is_unique, term, dbxref) = self.client._is_cvterm_unique(cv.CvTerm(
            cv_id=self.default_cv.cv_id, dbxref_id=self.default_dbxref.dbxref_id+1, name="testterm", is_obsolete=0))
        self.assertFalse(is_unique)

        # Test a CV term with different properties - should be unique
        (is_unique, term, dbxref) = self.client._is_cvterm_unique(cv.CvTerm(
            cv_id=self.default_cv.cv_id, dbxref_id=self.default_dbxref.dbxref_id+1, name="otherterm", is_obsolete=1))
        self.assertTrue(is_unique)

    def test_mark_cvterm_as_obsolete(self):
        # Tests the functionality that marks a CV term as obsolete

        # Mark a previously non-obsolete CV term as obsolete
        marked = self.client._mark_cvterm_as_obsolete(self.default_cvterm)
        self.assertTrue(marked)
        self.assertEqual(self.default_cvterm.is_obsolete, 1)
        self.assertEqual(self.default_cvterm.name, "obsolete testterm")

        # Try to mark an already obsolete CV term as obsolete
        marked = self.client._mark_cvterm_as_obsolete(self.default_cvterm)
        self.assertFalse(marked)
        self.assertEqual(self.default_cvterm.is_obsolete, 1)
        self.assertEqual(self.default_cvterm.name, "obsolete testterm")

        # Enforce marking a CV term as "even more obsolete" by testing an obsolete CV term with the same properties
        # as a term existing in the database
        further_dbxref = general.DbxRef(db_id=self.default_db.db_id, accession="furtheraccession")
        self.client.add_and_flush(further_dbxref)
        further_cvterm = cv.CvTerm(cv_id=self.default_cvterm.cv_id, dbxref_id=further_dbxref.dbxref_id,
                                   name=self.default_cvterm.name, is_obsolete=1)
        marked = self.client._mark_cvterm_as_obsolete(further_cvterm)
        self.assertTrue(marked)
        self.assertEqual(further_cvterm.is_obsolete, 2)
        self.assertEqual(further_cvterm.name, "obsolete testterm")

    def test_handle_cvterm(self):
        # Tests the insertion and update of cvterm entries

        # Populate the database with essentials
        ontology_terms = {"defaultdb:001": pronto.Term("defaultdb:001", "term1"),
                          "defaultdb:002": pronto.Term("defaultdb:002", "term2")}

        # Call the function with a dbxref that is not part of the ontology, and check that the corresponding CV term
        # is marked as obsolete
        obsolete_cvterm = self.client._handle_cvterms(ontology_terms, self.default_db, self.default_dbxref,
                                                      self.default_cv.cv_id)
        self.assertEqual(obsolete_cvterm.is_obsolete, 1)
        self.assertEqual(self.client._cvterm_inserts, 0)
        self.assertEqual(self.client._cvterm_updates, 0)
        self.assertEqual(self.client._cvterm_deletes, 1)

        # Call the function with a dbxref that is part of the ontology, and check that a CV term is created
        novel_dbxref = general.DbxRef(db_id=self.default_db.db_id, accession="001")
        self.client.add_and_flush(novel_dbxref)
        novel_cvterm = self.client._handle_cvterms(ontology_terms, self.default_db, novel_dbxref, self.default_cv.cv_id)
        self.assertIsNotNone(novel_cvterm.cvterm_id)
        self.assertEqual(novel_cvterm.dbxref_id, novel_dbxref.dbxref_id)
        self.assertEqual(novel_cvterm.name, "term1")
        self.assertEqual(self.client._cvterm_inserts, 1)
        self.assertEqual(self.client._cvterm_updates, 0)
        self.assertEqual(self.client._cvterm_deletes, 1)

        # Call the function with changed ontology term, and check that this updates the existing entry
        ontology_terms["defaultdb:001"].name = "newname"
        updated_cvterm = self.client._handle_cvterms(ontology_terms, self.default_db, novel_dbxref,
                                                     self.default_cv.cv_id)
        self.assertEqual(novel_cvterm.cvterm_id, updated_cvterm.cvterm_id)
        self.assertEqual(updated_cvterm.name, "newname")
        self.assertEqual(self.client._cvterm_inserts, 1)
        self.assertEqual(self.client._cvterm_updates, 1)
        self.assertEqual(self.client._cvterm_deletes, 1)

        # Call the function with a dbxref that corresponds to an existing CV term, and check that nothing happens
        unaltered_cvterm = self.client._handle_cvterms(ontology_terms, self.default_db, novel_dbxref,
                                                       self.default_cv.cv_id)
        self.assertEqual(unaltered_cvterm, updated_cvterm)
        self.assertEqual(self.client._cvterm_inserts, 1)
        self.assertEqual(self.client._cvterm_updates, 1)
        self.assertEqual(self.client._cvterm_deletes, 1)

        # Call the function with an obsolete ontology term that has the same properties as an existing CV term, and
        # check that the new CV term is marked as 'more obsolete' before insertion
        ontology_terms["defaultdb:003"] = pronto.Term("defaultdb:003", "testterm", other={"is_obsolete": ["True"]})
        duplicate_obsolete_dbxref = general.DbxRef(db_id=self.default_db.db_id, accession="003")
        self.client.add_and_flush(duplicate_obsolete_dbxref)
        duplicate_obsolete_cvterm = self.client._handle_cvterms(ontology_terms, self.default_db,
                                                                duplicate_obsolete_dbxref, self.default_cv.cv_id)
        self.assertEqual(obsolete_cvterm.is_obsolete, 1)
        self.assertEqual(duplicate_obsolete_cvterm.is_obsolete, 2)
        self.assertEqual(self.client._cvterm_inserts, 2)
        self.assertEqual(self.client._cvterm_updates, 1)
        self.assertEqual(self.client._cvterm_deletes, 1)

        # Call the function with a non-obsolete ontology term that has the same properties as an existing CV term, and
        # check that the existing term is updated first
        ontology_terms["defaultdb:001"].name = "oldname"
        ontology_terms["defaultdb:004"] = pronto.Term("defaultdb:004", "newname")
        duplicate_dbxref = general.DbxRef(db_id=self.default_db.db_id, accession="004")
        self.client.add_and_flush(duplicate_dbxref)
        duplicate_cvterm = self.client._handle_cvterms(ontology_terms, self.default_db, duplicate_dbxref,
                                                       self.default_cv.cv_id)
        self.assertEqual(novel_cvterm.name, "oldname")
        self.assertEqual(duplicate_cvterm.name, "newname")
        self.assertEqual(self.client._cvterm_inserts, 3)
        self.assertEqual(self.client._cvterm_updates, 2)
        self.assertEqual(self.client._cvterm_deletes, 1)

    def test_handle_typedef(self):
        # Tests the insertion of relationship-type CV terms

        # Insert a new entry
        relationship = pronto.Relationship("new_relationship")
        first_typedef = self.client._handle_typedef(relationship, self.default_db, self.default_cv)
        self.assertEqual(first_typedef.name, "new_relationship")
        self.assertEqual(first_typedef.cv_id, self.default_cv.cv_id)
        self.assertTrue(first_typedef.is_relationshiptype)
        self.assertIsNotNone(first_typedef.cvterm_id)

        # Insert a further entry with a different name
        relationship = pronto.Relationship("another_relationship")
        second_typedef = self.client._handle_typedef(relationship, self.default_db, self.default_cv)
        self.assertNotEqual(first_typedef.cvterm_id, second_typedef.cvterm_id)

        # Try to insert a further entry with the same authority, and check that the function returns the existing entry
        third_typedef = self.client._handle_typedef(relationship, self.default_db, self.default_cv)
        self.assertEqual(third_typedef.cvterm_id, second_typedef.cvterm_id)

    def test_handle_comments(self):
        # Tests the insertion, update and deletion of comments (cvtermprop entries)

        # Check that an ontology term without comment will not create a cvtermprop entry
        term = pronto.Term("defaultdb:defaultaccession")
        first_cvtermprop = self.client._handle_comments(term, self.default_cvterm)
        self.assertIsNone(first_cvtermprop)

        # Insert a comment
        term.other = {"comment": ["testcomment"]}
        second_cvtermprop = self.client._handle_comments(term, self.default_cvterm)
        self.assertIsNotNone(second_cvtermprop.cvtermprop_id)
        self.assertEqual(second_cvtermprop.cvterm_id, self.default_cvterm.cvterm_id)
        self.assertEqual(second_cvtermprop.type_id, self.client._comment_term.cvterm_id)
        self.assertEqual(second_cvtermprop.value, "testcomment")

        # Try to insert another comment and check that this updates the existing entry
        term.other = {"comment": ["othercomment"]}
        third_cvtermprop = self.client._handle_comments(term, self.default_cvterm)
        self.assertEqual(third_cvtermprop.cvtermprop_id, second_cvtermprop.cvtermprop_id)
        self.assertEqual(third_cvtermprop.value, "othercomment")

    def test_handle_synonyms(self):
        # Tests the insertion, update and deletion of synonyms of CV terms

        # Check that an ontology term without synonym will not create a cvtermsynonym entry
        term = pronto.Term("defaultdb:defaultaccession")
        first_synonyms = self.client._handle_synonyms(term, self.default_cvterm)
        self.assertEqual(len(first_synonyms), 0)

        # Insert a synonym
        term.synonyms = {pronto.Synonym("another_name", "EXACT")}
        second_synonyms = self.client._handle_synonyms(term, self.default_cvterm)
        self.assertEqual(len(second_synonyms), 1)
        self.assertIsNotNone(second_synonyms[0].cvtermsynonym_id)
        self.assertEqual(second_synonyms[0].synonym, "another_name")
        self.assertEqual(second_synonyms[0].type_id, self.client._synonym_type_terms["exact"].cvterm_id)

        # Try to insert an existing synonym with changed scope, and check that this updates the existing entry
        term.synonyms = {pronto.Synonym("another_name", "NARROW")}
        third_synonyms = self.client._handle_synonyms(term, self.default_cvterm)
        self.assertEqual(len(third_synonyms), 1)
        self.assertEqual(third_synonyms[0].cvtermsynonym_id, second_synonyms[0].cvtermsynonym_id)
        self.assertEqual(third_synonyms[0].type_id, self.client._synonym_type_terms["narrow"].cvterm_id)

        # Insert another synonym
        term.synonyms.add(pronto.Synonym("yet_another_name", "BROAD"))
        fourth_synonyms = self.client._handle_synonyms(term, self.default_cvterm)
        self.assertEqual(len(fourth_synonyms), 2)

    def test_handle_cross_references(self):
        # Tests the insertion, update and deletion of cross references for CV terms

        # Check that an ontology term without cross references will not create a cvterm_dbxref entry
        term = pronto.Term("defaultdb:defaultaccession")
        first_crossrefs = self.client._handle_cross_references(term, self.default_cvterm)
        self.assertEqual(len(first_crossrefs), 0)

        # Insert a further cross reference to the same database
        term.other = {"alt_id": ["defaultdb:furtheraccession"]}
        second_crossrefs = self.client._handle_cross_references(term, self.default_cvterm)
        self.assertEqual(len(second_crossrefs), 1)
        corresponding_dbxref = self.client.find_or_insert(general.DbxRef, db_id=self.default_db.db_id,
                                                          accession="furtheraccession")
        self.assertEqual(corresponding_dbxref.dbxref_id, second_crossrefs[0].dbxref_id)

        # Insert a cross reference to another database
        term.other = {"xref": ["otherdb:otheraccession 'with comment'"]}
        third_crossrefs = self.client._handle_cross_references(term, self.default_cvterm)
        self.assertEqual(len(third_crossrefs), 1)
        other_db = self.client.query_table(general.Db, name="otherdb").first()
        self.assertIsNotNone(other_db)
        corresponding_dbxref = self.client.find_or_insert(general.DbxRef, db_id=other_db.db_id,
                                                          accession="otheraccession")
        self.assertEqual(corresponding_dbxref.dbxref_id, third_crossrefs[0].dbxref_id)

    def test_handle_relationships(self):
        # Tests the insertion, update and deletion of relationships between CV terms

        # Populate the database with essentials
        other_dbxref = general.DbxRef(db_id=self.default_db.db_id, accession="otheraccession")
        self.client.add_and_flush(other_dbxref)
        other_cvterm = cv.CvTerm(cv_id=self.default_cv.cv_id, dbxref_id=other_dbxref.dbxref_id, name="otherterm")
        self.client.add_and_flush(other_cvterm)
        default_id = ontology.create_dbxref(self.default_db.name, self.default_dbxref.accession)
        other_id = ontology.create_dbxref(self.default_db.name, other_dbxref.accession)
        all_cvterms = {default_id: self.default_cvterm, other_id: other_cvterm}

        # Check that an ontology term without relationships will not create a cvterm_relationship entry
        term = pronto.Term("defaultdb:defaultaccession")
        first_relationships = self.client._handle_relationships(term, self.default_cvterm, all_cvterms)
        self.assertEqual(len(first_relationships), 0)

        # Insert a relationship
        term.relations = {pronto.Relationship("part_of"): [pronto.Term(other_id, other_cvterm.name)]}
        second_relationships = self.client._handle_relationships(term, self.default_cvterm, all_cvterms)
        self.assertEqual(len(second_relationships), 1)
        self.assertIsNotNone(second_relationships[0].cvterm_relationship_id)
        self.assertEqual(second_relationships[0].subject_id, self.default_cvterm.cvterm_id)
        self.assertEqual(second_relationships[0].object_id, other_cvterm.cvterm_id)
        self.assertEqual(second_relationships[0].type_id, self.client._relationship_terms["part_of"].cvterm_id)

        # Insert a new relationship between the same terms
        term.relations.clear()
        term.relations = {pronto.Relationship("is_a"): [pronto.Term(other_id, other_cvterm.name)]}
        third_relationships = self.client._handle_relationships(term, self.default_cvterm, all_cvterms)
        self.assertEqual(len(third_relationships), 1)
        self.assertEqual(third_relationships[0].type_id, self.client._relationship_terms["is_a"].cvterm_id)

    def test_mark_obsolete_terms(self):
        # Populate the database with essentials
        other_dbxref = general.DbxRef(db_id=self.default_db.db_id, accession="otheraccession")
        self.client.add_and_flush(other_dbxref)
        other_cvterm = cv.CvTerm(cv_id=self.default_cv.cv_id, dbxref_id=other_dbxref.dbxref_id, name="otherterm")
        self.client.add_and_flush(other_cvterm)

        # Check that the two inserted CV terms are marked as obsolete
        terms = {}
        marked_cvterms = self.client._mark_obsolete_terms(terms, self.default_db)
        self.assertEqual(len(marked_cvterms), 2)
        self.assertIn(marked_cvterms[0].cvterm_id, [self.default_cvterm.cvterm_id, other_cvterm.cvterm_id])
        self.assertIn(marked_cvterms[1].cvterm_id, [self.default_cvterm.cvterm_id, other_cvterm.cvterm_id])


if __name__ == '__main__':
    unittest.main(verbosity=2, buffer=True)



from ..io import iobase
from ..orm import general, cv, pub


class EssentialsClient(iobase.ChadoClient):
    """Class for importing essential basic entries into the Chado tables"""
    
    def load(self):
        """Import essential basic entries into various database tables"""
        self._load_generic_entries()
        self._load_relationship_entries()
        self._load_synonymtype_entries()
        self._load_cvterm_property_type_entries()
        self._load_feature_property_entries()
        self._load_genedb_synonymtype_entries()
        self._load_genedb_misc_entries()
        self._load_genedb_products_vocabulary()
        self.session.commit()

    def _load_generic_entries(self):
        """Import generic entries in the db, dbxref, cv, cvterm, and pub tables"""
        new_generic_db = general.Db(name="null", description="a fake database for locally created essentials")
        generic_db = self._handle_db(new_generic_db)
        new_generic_dbxref = general.DbxRef(db_id=generic_db.db_id, accession="null")
        generic_dbxref = self._handle_dbxref(new_generic_dbxref, generic_db.name)
        new_generic_cv = cv.Cv(name="null", definition="a fake vocabulary for locally created essentials")
        generic_cv = self._handle_cv(new_generic_cv)
        new_generic_cvterm = cv.CvTerm(cv_id=generic_cv.cv_id, dbxref_id=generic_dbxref.dbxref_id, name="null")
        generic_cvterm = self._handle_cvterm(new_generic_cvterm, generic_cv.name)
        new_generic_pub = pub.Pub(uniquename="null", type_id=generic_cvterm.cvterm_id)
        self._handle_pub(new_generic_pub)

    def _load_relationship_entries(self, *further_terms):
        """Import CV terms for relationships
        (note: all other basic relationship terms are part of the RO relationship ontology)"""
        new_relationship_db = general.Db(name="RO")
        relationship_db = self._handle_db(new_relationship_db)
        new_relationship_cv = cv.Cv(name="relationship")
        relationship_cv = self._handle_cv(new_relationship_cv)

        terms = ["is_a"] + [term for term in further_terms]
        for term in terms:

            new_dbxref = general.DbxRef(db_id=relationship_db.db_id, accession=term)
            dbxref = self._handle_dbxref(new_dbxref, relationship_db.name)
            new_cvterm = cv.CvTerm(cv_id=relationship_cv.cv_id, dbxref_id=dbxref.dbxref_id, name=term,
                                   is_relationshiptype=1)
            self._handle_cvterm(new_cvterm, relationship_cv.name)

    def _load_synonymtype_entries(self):
        """Import CV terms for synonym types"""
        new_synonymtype_db = general.Db(name="internal")
        synonymtype_db = self._handle_db(new_synonymtype_db)
        new_synonymtype_cv = cv.Cv(name="synonym_type")
        synonymtype_cv = self._handle_cv(new_synonymtype_cv)

        for term in ["exact", "narrow", "broad", "related"]:

            new_dbxref = general.DbxRef(db_id=synonymtype_db.db_id, accession=term)
            dbxref = self._handle_dbxref(new_dbxref, synonymtype_db.name)
            new_cvterm = cv.CvTerm(cv_id=synonymtype_cv.cv_id, dbxref_id=dbxref.dbxref_id, name=term)
            self._handle_cvterm(new_cvterm, synonymtype_cv.name)

    def _load_cvterm_property_type_entries(self):
        """Import CV terms for property types"""
        new_propertytype_db = general.Db(name="internal")
        propertytype_db = self._handle_db(new_propertytype_db)
        new_propertytype_cv = cv.Cv(name="cvterm_property_type")
        propertytype_cv = self._handle_cv(new_propertytype_cv)

        # non-relationship types
        for term in ["comment", "is_anonymous", "is_transitive", "is_anti_symmetric", "is_reflexive", "is_symmetric",
                     "is_cyclic"]:

            new_dbxref = general.DbxRef(db_id=propertytype_db.db_id, accession=term)
            dbxref = self._handle_dbxref(new_dbxref, propertytype_db.name)
            new_cvterm = cv.CvTerm(cv_id=propertytype_cv.cv_id, dbxref_id=dbxref.dbxref_id, name=term)
            self._handle_cvterm(new_cvterm, propertytype_cv.name)

        # relationship types
        for term in ["intersection_of", "disjoint_from"]:

            new_dbxref = general.DbxRef(db_id=propertytype_db.db_id, accession=term)
            dbxref = self._handle_dbxref(new_dbxref, propertytype_db.name)
            new_cvterm = cv.CvTerm(cv_id=propertytype_cv.cv_id, dbxref_id=dbxref.dbxref_id, name=term,
                                   is_relationshiptype=1)
            self._handle_cvterm(new_cvterm, propertytype_cv.name)

    def _load_feature_property_entries(self):
        """Import CV terms for feature properties (note: these terms are also part of the SOFP ontology)"""
        new_feature_property_db = general.Db(name="SOFP")
        feature_property_db = self._handle_db(new_feature_property_db)
        new_feature_property_cv = cv.Cv(name="feature_property")
        feature_property_cv = self._handle_cv(new_feature_property_cv)

        for term in ["comment", "score", "source", "description", "date"]:

            new_dbxref = general.DbxRef(db_id=feature_property_db.db_id, accession=term)
            dbxref = self._handle_dbxref(new_dbxref, feature_property_db.name)
            new_cvterm = cv.CvTerm(cv_id=feature_property_cv.cv_id, dbxref_id=dbxref.dbxref_id, name=term)
            self._handle_cvterm(new_cvterm, feature_property_cv.name)

    def _load_genedb_synonymtype_entries(self):
        """Import CV terms for specific GeneDB synonym types"""
        new_synonymtype_db = general.Db(name="genedb_misc")
        synonymtype_db = self._handle_db(new_synonymtype_db)
        new_synonymtype_cv = cv.Cv(name="genedb_synonym_type")
        synonymtype_cv = self._handle_cv(new_synonymtype_cv)

        for term in ["synonym", "alias", "systematic_id", "previous_systematic_id", "temporary_systematic_id"]:

            new_dbxref = general.DbxRef(db_id=synonymtype_db.db_id, accession=term)
            dbxref = self._handle_dbxref(new_dbxref, synonymtype_db.name)
            new_cvterm = cv.CvTerm(cv_id=synonymtype_cv.cv_id, dbxref_id=dbxref.dbxref_id, name=term)
            self._handle_cvterm(new_cvterm, synonymtype_cv.name)

    def _load_genedb_misc_entries(self):
        """Import specific CV terms for GeneDB"""
        new_misc_db = general.Db(name="genedb_misc")
        misc_db = self._handle_db(new_misc_db)
        new_misc_cv = cv.Cv(name="genedb_misc")
        misc_cv = self._handle_cv(new_misc_cv)

        for term in ["top_level_seq", "evidence", "genedb_public"]:

            new_dbxref = general.DbxRef(db_id=misc_db.db_id, accession=term)
            dbxref = self._handle_dbxref(new_dbxref, misc_db.name)
            new_cvterm = cv.CvTerm(cv_id=misc_cv.cv_id, dbxref_id=dbxref.dbxref_id, name=term)
            self._handle_cvterm(new_cvterm, misc_cv.name)

    def _load_genedb_products_vocabulary(self):
        """Import CV for GeneDB products"""
        new_product_db = general.Db(name="PRODUCT")
        self._handle_db(new_product_db)
        new_product_cv = cv.Cv(name="genedb_products")
        self._handle_cv(new_product_cv)

    def _load_sequence_type_entries(self):
        """Import CV terms for sequence types; for testing only (all terms are part of the SO sequence ontology)"""
        new_sequence_db = general.Db(name="SO")
        sequence_db = self._handle_db(new_sequence_db)
        new_sequence_cv = cv.Cv(name="sequence")
        sequence_cv = self._handle_cv(new_sequence_cv)

        for term in ["gene", "intron", "exon", "CDS", "mRNA", "chromosome", "contig", "supercontig", "region"]:

            new_dbxref = general.DbxRef(db_id=sequence_db.db_id, accession=term)
            dbxref = self._handle_dbxref(new_dbxref, sequence_db.name)
            new_cvterm = cv.CvTerm(cv_id=sequence_cv.cv_id, dbxref_id=dbxref.dbxref_id, name=term)
            self._handle_cvterm(new_cvterm, sequence_cv.name)

        self.session.commit()

    def _load_further_relationship_entries(self):
        """Import further CV terms for relationships; for testing only"""
        self._load_relationship_entries("part_of", "derives_from")
        self.session.commit()
